import logging
from datetime import timedelta
from typing import Callable

import rx.operators as ops
from rx.core.abc import Disposable
from rx.disposable import CompositeDisposable
from rx.subject import Subject

from _add_lib_to_path import add_lib_to_path

# This function adds the project root to PATH
add_lib_to_path()

from sfm2.interface.sample_types import VectorSample
from sfm2.interface.sfm2_init_options import Sfm2CollectionInitOptions
from sfm2.utils.time_sync import TimeSync
from sfm2.sfm2_factory import Sfm2Factory

_THRESHOLD_G = 0.5
_COOLDOWN_SAMPLES = 800
_SEPARATOR_TIMEOUT_S = 2

_OPTIONS = Sfm2CollectionInitOptions(
    allow_duplicate_keys=False,
    key_by_id_not_name=True,
)


def main():
    logging.basicConfig(level=logging.WARNING)

    print("This sample can be used to test the time synchronization between multiple connected SFM2s.\n"
          "The script will synchronize clocks of the SFM2 and enable acceleration spikes detection on each of them.\n"
          f"Each SFM2 will report the exact time it detects acceleration above the threshold of {_THRESHOLD_G}g\n"
          "\n"
          "To conduct this test lay a few SFM2s flat on your desk. Run this script and wait until the clocks are synchronized.\n"
          "Then tap the desk so that all the SFM2s detects the shock. A line will be printed for each device.\n"
          "If necessary try striking harder or lowering the threshold acceleration.\n"
          "If that doesn't help try stacking the SFM2s on top of each other and tap on the top.\n"
          "\n"
          "The SFM2s are sampling at 833Hz which limits the precision of this test to a few milliseconds.\n"
          "\n")

    disposables = CompositeDisposable()

    all_sfm2s = Sfm2Factory.connect_all_found(collection_options=_OPTIONS)
    print(f"Connected to {len(all_sfm2s)} SFM2s")

    sync = TimeSync()
    sync_status = sync.sync(all_sfm2s)
    disposables.add(sync_status)

    separator_printer = _SeparatorPrinter()

    for sfm2 in all_sfm2s:
        sfm2.sensor_fusion.rate = 800

    print("Waiting for sync...")
    sync_status.wait_for_sync(timeout=timedelta(seconds=20))
    print("Synchronized.")

    for sfm2 in all_sfm2s:
        processor = _Processor(sfm2.name, separator_printer.print)
        d = sfm2.sensor_fusion.linear_acceleration.subscribe(processor.on_la)
        disposables.add(d)

    print("All configured. Try tapping...")
    input("Click enter twice to exit\n")
    disposables.dispose()


class _Processor:
    def __init__(self, name: str, print_separator: Callable[[], None]):
        self._name = name
        self._print_separator = print_separator

        self._samples_since_above = 0
        self._last_la = None

    def on_la(self, la: VectorSample) -> None:
        if la.magnitude > _THRESHOLD_G:
            if self._samples_since_above > _COOLDOWN_SAMPLES:
                self._process_cross(la)

            self._samples_since_above = 0
        else:
            self._samples_since_above += 1

        self._last_la = la

    def _process_cross(self, la: VectorSample):
        cross_time = self._interpolate_cross_time_s(la)
        print(f"{self._name}: {cross_time:.4f}s")
        self._print_separator()

    def _interpolate_cross_time_s(self, la_above_th: VectorSample) -> float:
        t1 = self._last_la.timestamp.total_seconds()
        t2 = la_above_th.timestamp.total_seconds()
        d1 = _THRESHOLD_G - self._last_la.magnitude
        d2 = la_above_th.magnitude - _THRESHOLD_G
        return (d1 * t2 + d2 * t1) / (d1 + d2)


class _SeparatorPrinter(Disposable):
    def __init__(self):
        self._subject = Subject()
        self._disposable = self._subject.pipe(
            ops.debounce(2)
        ).subscribe(lambda _: print("==="))

    def print(self):
        self._subject.on_next(True)

    def dispose(self):
        self._disposable.dispose()
        self._subject.dispose()


if __name__ == "__main__":
    main()
