import logging
from typing import Optional, Tuple

import rx
from rx.core.abc import Disposable
from rx.disposable import CompositeDisposable

from _add_lib_to_path import add_lib_to_path

# This function adds the project root to PATH
add_lib_to_path()

from sfm2.interface.found_sfm2 import FoundSfm2
from sfm2.interface.sfm2 import Sfm2
from sfm2.sfm2_factory import Sfm2Factory
from sfm2.utils.throughput import calculate_throughput
from sfm2.utils.csv_writer import write_to_csv, CsvWriterOptions


def main():
    logging.basicConfig(level=logging.WARNING)

    found_sfm2s = Sfm2Factory.find_all_on_serial_ports()

    sfm2s = [connect_noexcept(f) for f in found_sfm2s]

    if len(sfm2s) == 0:
        print("Could not connect to any SFM2")
        exit(-1)

    disposables = CompositeDisposable()

    for s, l in sfm2s:
        d = setup_csv_logging(s, l)
        disposables.add(d)

    print("Setup completed. Enabling data streams...")

    for s, _ in sfm2s:
        s.acc.rate = 1600
        s.gyr.rate = 1600
        s.sensor_fusion.rate = 400

    input("Click enter twice to exit\n")
    print("Disposing...")
    disposables.dispose()

    for s, _ in sfm2s:
        s.reset()


def connect_noexcept(found: FoundSfm2) -> Optional[Tuple[Sfm2, str]]:
    try:
        print(f"Connecting to {found.label}...")
        connected = found.connect()
        connected.reset()
        print("\tConnected")
        return connected, found.label
    except Exception as e:
        print(f"\tError: {e}")
        return None


def setup_csv_logging(sfm2: Sfm2, device_label: str) -> Disposable:
    # This option disables creating new thread for each CSV file.
    # It may improve performance when you try to log many streams.
    options = CsvWriterOptions(create_new_thread=False)

    d1 = write_to_csv(sfm2.acc.data, f"{device_label}_AD.csv", options)
    d2 = write_to_csv(sfm2.gyr.data, f"{device_label}_GD.csv", options)
    d3 = write_to_csv(sfm2.sensor_fusion.quaternion, f"{device_label}_SFQT.csv", options)
    d4 = write_to_csv(sfm2.sensor_fusion.linear_acceleration, f"{device_label}_SFLA.csv", options)

    # You don't need to calculate the throughput. It's just useful for monitoring the data.
    d5 = setup_throughput_logging(sfm2, device_label)

    return CompositeDisposable(d1, d2, d3, d4, d5)


def setup_throughput_logging(sfm2: Sfm2, label: str) -> Disposable:
    interval_s = 3
    acc_throughput = calculate_throughput(sfm2.acc.data, interval_s=interval_s)
    gyr_throughput = calculate_throughput(sfm2.gyr.data, interval_s=interval_s)
    sfqt_throughput = calculate_throughput(sfm2.sensor_fusion.quaternion, interval_s=interval_s)
    sfla_throughput = calculate_throughput(sfm2.sensor_fusion.linear_acceleration, interval_s=interval_s)

    return rx.zip(acc_throughput, gyr_throughput, sfqt_throughput, sfla_throughput) \
        .subscribe(lambda th: print(f"{label} AD:{th[0]} GD:{th[1]} SFQT:{th[2]} SFLA:{th[3]}"))


if __name__ == '__main__':
    main()
