from threading import RLock

from rx import operators as ops
from rx.core.typing import Observable

from sfm2._implementation.commands.command import Command
from sfm2._implementation.commands.wait_for_response import WaitForResponse
from sfm2._implementation.serial_port.serial_port import SerialPort
from sfm2.interface.exceptions import CommandResponseValueMismatch, CommandFailedException
from sfm2.interface.sfm2_init_options import Sfm2InitOptions

_CMD_OK = "OK"
_CMD_FAIL = "FAIL"


class CommandInterface:
    def __init__(self, port: SerialPort, lock: RLock, options: Sfm2InitOptions):
        self._lock = lock
        self._options = options
        self._port = port

    def query(self, cmd: Command, check_if_failed: bool = True) -> str:
        with self._lock:
            with self._create_wait(cmd) as w:
                self._port.send(cmd.query_str)
                response = w()
                self._check_if_failed(cmd, response, check_if_failed)
                return response

    def set_value(self, cmd: Command, value: str, check_if_failed: bool = True) -> str:
        with self._lock:
            with self._create_wait(cmd) as w:
                line = cmd.set_value_str(value)
                self._port.send(line)
                response = w()
                self._check_if_failed(cmd, response, check_if_failed)
                return response

    def set_value_exact(self, cmd: Command, value: str) -> None:
        response = self.set_value(cmd, value)
        if response != value:
            CommandResponseValueMismatch(
                command_name=cmd.name,
                got_value=response,
                set_value=value
            ).raise_and_log_warning()

    def command(self, cmd: Command, check_if_failed: bool = True) -> str:
        with self._lock:
            with self._create_wait(cmd) as w:
                line = cmd.command_str
                self._port.send(line)
                response = w()
                self._check_if_failed(cmd, response, check_if_failed)
                return response

    def command_and_check_if_ok(self, cmd: Command) -> None:
        with self._lock:
            with self._create_wait(cmd) as w:
                line = cmd.command_str
                self._port.send(line)
                response = w()
                if response != _CMD_OK:
                    CommandFailedException(cmd.name).raise_and_log_warning()

    def observe_values(self, cmd: Command) -> Observable[str]:
        return self._port.incoming_text.pipe(
            ops.filter(lambda l: l.startswith(cmd.name)),
            ops.map(lambda l: cmd.strip_value(l))
        )

    def _create_wait(self, cmd: Command) -> WaitForResponse:
        return WaitForResponse(self._port, cmd, self._options.response_timeout_s)

    @staticmethod
    def _check_if_failed(cmd: Command, response: str, enabled: bool) -> None:
        if enabled and response == _CMD_FAIL:
            CommandFailedException(cmd.name).raise_and_log_warning()
