from __future__ import annotations

import logging
from typing import Optional, Callable

from rx import operators as ops
from rx.core import ConnectableObservable
from rx.core.abc import Disposable

from sfm2._implementation.commands.command import Command
from sfm2._implementation.internal_utils.wrap_exception import wrap_exception
from sfm2._implementation.serial_port.serial_port import SerialPort
from sfm2.interface.exceptions import CommandResponseTimeoutException

_log = logging.getLogger(__name__)


class WaitForResponse:
    def __init__(self, port: SerialPort, cmd: Command, timeout_s: float):
        self._cmd = cmd
        self._timeout_s: float = timeout_s
        self._disposable: Optional[Disposable] = None
        self._response: ConnectableObservable[str] = port.incoming_text.pipe(
            ops.filter(lambda l: l.startswith(cmd.name)),
            ops.take(1),
            ops.map(lambda l: cmd.strip_value(l)),
            ops.replay(buffer_size=1)
        )

    def __enter__(self) -> Callable[[], str]:
        self._disposable = self._response.connect()
        return self._wait

    def __exit__(self, typ, value, traceback):
        if self._disposable is not None:
            self._disposable.dispose()

    def _create_timeout(self):
        return CommandResponseTimeoutException(
            command_name=self._cmd.name,
            timeout_s=self._timeout_s)

    @wrap_exception(create_wrapping_exc=_create_timeout)
    def _wait(self) -> str:
        result = self._response.pipe(
            ops.timeout(self._timeout_s)
        ).run()
        return result
