import logging
from threading import RLock
from typing import Callable, Dict, Iterable

import rx.core.abc
import rx.disposable

import sfm2._implementation.commands.command_list as cmds
from sfm2._implementation.commands.command import Command
from sfm2._implementation.commands.command_interface import CommandInterface
from sfm2.interface.exceptions import Sfm2InternalException

_log = logging.getLogger(__name__)


class DataEnabler:
    def __init__(self, cmd: CommandInterface):
        self._lock = RLock()
        self._cmd = cmd
        self._enabled_streams: _EnabledStreams = _EnabledStreams()

    def create_enable(self, cmd: Command) -> Callable[[], rx.core.abc.Disposable]:
        def disable():
            self._disable_data(cmd)

        def enable():
            self._enable_data(cmd)
            return rx.disposable.Disposable(disable)

        return enable

    def reenable(self):
        if self._enabled_streams.empty():
            return

        _log.debug("Reenabling data streams")
        self._turn_on(cmds.binmode)
        for cmd in self._enabled_streams:
            _log.debug(f"Reenabling {cmd}")
            self._turn_on(cmd)

    def _enable_data(self, cmd: Command) -> None:
        with self._lock:
            if cmd not in self._enabled_streams:
                self._enable_binary_mode_if_necessary()
                self._turn_on(cmd)
            self._enabled_streams.add(cmd)

    def _disable_data(self, cmd: Command) -> None:
        with self._lock:
            self._enabled_streams.remove(cmd)
            if cmd not in self._enabled_streams:
                self._turn_off(cmd)
                self._disable_binary_mode_if_necessary()

    def _enable_binary_mode_if_necessary(self) -> None:
        if self._enabled_streams.empty():
            self._turn_on(cmds.binmode)

    def _disable_binary_mode_if_necessary(self) -> None:
        if self._enabled_streams.empty():
            self._turn_off(cmds.binmode)

    def _turn_on(self, cmd: Command) -> None:
        _log.info(f"Enabling {cmd.name}")
        self._cmd.set_value_exact(cmd, "1")

    def _turn_off(self, cmd: Command) -> None:
        _log.info(f"Disabling {cmd.name}")
        self._cmd.set_value_exact(cmd, "0")


class _EnabledStreams:
    def __init__(self):
        self._dict: Dict[Command, int] = dict()

    def __contains__(self, cmd: Command) -> bool:
        return self._get_count(cmd) > 0

    def __iter__(self) -> Iterable[Command]:
        return iter(k for k, v in self._dict.items() if v > 0)

    def empty(self) -> bool:
        return not self._dict

    def any(self) -> bool:
        return not self.empty()

    def add(self, cmd: Command) -> None:
        count = self._get_count(cmd)
        self._dict[cmd] = count + 1

    def remove(self, cmd: Command) -> None:
        count = self._get_count(cmd)
        if count == 0:
            Sfm2InternalException(f"Tried removing stream {cmd} but there isn't any").raise_and_log_error(log=_log)

        if count == 1:
            del self._dict[cmd]
        else:
            self._dict[cmd] = count - 1

    def _get_count(self, cmd: Command) -> int:
        existing = self._dict.get(cmd)
        return existing or 0
