import logging
from datetime import timedelta
from typing import Any, Union

import sfm2._implementation.commands.command_list as cmds
from sfm2._implementation.commands.command import Command
from sfm2._implementation.commands.command_interface import CommandInterface
from sfm2._implementation.internal_utils.find_sample_rate import find_sample_rate
from sfm2.interface.env_sensor_config_types import PressureSampleRate, TemperatureSamplingMode
from sfm2.interface.sfm2_config import Sfm2Config, SampleRate, AccelFullRange, GyroFullRange
from sfm2.interface.sfm2_init_options import Sfm2InitOptions

_log = logging.getLogger(__name__)


class Sfm2ConfigImpl(Sfm2Config):
    def __init__(self, cmd: CommandInterface, options: Sfm2InitOptions):
        self._cmd = cmd
        self._options = options

    @property
    def asr(self) -> SampleRate:
        return self._read_sample_rate(cmds.asr)

    @asr.setter
    def asr(self, rate: Union[SampleRate, float]) -> None:
        found_rate = find_sample_rate(rate)
        _log.info(f"Setting ASR to {found_rate} (requested {rate})")
        self._set_and_optionally_check(cmds.asr, found_rate.value)

    @property
    def afr(self) -> AccelFullRange:
        value_str = self._cmd.query(cmds.afr)
        value_int = int(value_str)
        return AccelFullRange(value_int)

    @afr.setter
    def afr(self, value: AccelFullRange) -> None:
        _log.info(f"Setting AFR to {value}")
        self._set_and_optionally_check(cmds.afr, value.value)

    @property
    def gsr(self) -> SampleRate:
        return self._read_sample_rate(cmds.gsr)

    @gsr.setter
    def gsr(self, rate: Union[SampleRate, float]) -> None:
        found_rate = find_sample_rate(rate)
        _log.info(f"Setting GSR to {found_rate} (requested {rate})")
        self._set_and_optionally_check(cmds.gsr, found_rate.value)

    @property
    def gfr(self) -> GyroFullRange:
        value_str = self._cmd.query(cmds.gfr)
        value_int = int(value_str)
        return GyroFullRange(value_int)

    @gfr.setter
    def gfr(self, value: GyroFullRange) -> None:
        _log.info(f"Setting GFR to {value}")
        self._set_and_optionally_check(cmds.gfr, value.value)

    @property
    def msr(self) -> SampleRate:
        return self._read_sample_rate(cmds.msr)

    @msr.setter
    def msr(self, rate: Union[SampleRate, float]) -> None:
        found_rate = find_sample_rate(rate, maximum=SampleRate.HZ_104)
        _log.info(f"Setting MSR to {found_rate} (requested {rate})")
        self._set_and_optionally_check(cmds.msr, found_rate.value)

    @property
    def sfor(self) -> SampleRate:
        return self._read_sample_rate(cmds.sfor)

    @sfor.setter
    def sfor(self, rate: Union[SampleRate, float]) -> None:
        found_rate = find_sample_rate(rate, maximum=SampleRate.HZ_833)
        _log.info(f"Setting SFOR to {found_rate} (requested {rate})")
        self._set_and_optionally_check(cmds.sfor, found_rate.value)

    @property
    def sfop(self) -> int:
        value = self._cmd.query(cmds.sfop)
        return int(value)

    @sfop.setter
    def sfop(self, prescaler: int) -> None:
        _log.info(f"Setting SFOP to {prescaler}")
        self._set_and_optionally_check(cmds.sfop, prescaler)

    @property
    def psr(self) -> PressureSampleRate:
        value = self._cmd.query(cmds.psr)
        value_int = int(value)
        return PressureSampleRate(value_int)

    @psr.setter
    def psr(self, rate: Union[PressureSampleRate, int]) -> None:
        rate_not_int = PressureSampleRate(rate)
        _log.info(f"Setting PSR to {rate_not_int}. Requested: {rate}")
        self._set_and_optionally_check(cmds.psr, rate_not_int.value)

    @property
    def altzero(self) -> float:
        value = self._cmd.query(cmds.altzero)
        return float(value)

    @altzero.setter
    def altzero(self, zero_altitude_pressure_hpa: float) -> None:
        _log.info(f"Setting ALTZERO to {zero_altitude_pressure_hpa}")
        as_float = float(zero_altitude_pressure_hpa)
        as_str = str(as_float)
        self._cmd.set_value(cmds.altzero, as_str)

    @property
    def pln(self) -> bool:
        return self._read_bool(cmds.pln)

    @pln.setter
    def pln(self, low_noise_enabled: bool) -> None:
        _log.info(f"Setting PLN to {low_noise_enabled}")
        self._set_bool(cmds.pln, low_noise_enabled)

    @property
    def tint(self) -> timedelta:
        value_str = self._cmd.query(cmds.tint)
        value_ms = int(value_str)
        return timedelta(milliseconds=value_ms)

    @tint.setter
    def tint(self, sampling_interval: timedelta) -> None:
        _log.info(f"Setting TINT to {sampling_interval}")
        seconds = sampling_interval.total_seconds()
        ms = int(seconds * 1000)
        self._set_and_optionally_check(cmds.tint, ms)

    @property
    def tmode(self) -> TemperatureSamplingMode:
        value_int = self._read_int(cmds.tmode)
        return TemperatureSamplingMode(value_int)

    @tmode.setter
    def tmode(self, mode: TemperatureSamplingMode) -> None:
        self._set_int(cmds.tmode, mode.value)

    def _set_and_optionally_check(self, cmd: Command, value: Any) -> None:
        value_str = str(value)
        if self._options.verify_all_cmd_response_values:
            self._cmd.set_value_exact(cmd, value_str)
        else:
            self._cmd.set_value(cmd, value_str)

    def _read_sample_rate(self, cmd: Command) -> SampleRate:
        value_str = self._cmd.query(cmd)
        value_int = int(value_str)
        return SampleRate(value_int)

    def _read_bool(self, cmd: Command) -> bool:
        value_int = self._read_int(cmd)
        return bool(value_int)

    def _set_bool(self, cmd: Command, value: bool) -> None:
        self._set_int(cmd, int(value))

    def _read_int(self, cmd: Command) -> int:
        value_str = self._cmd.query(cmd)
        return int(value_str)

    def _set_int(self, cmd: Command, value: int) -> None:
        self._set_and_optionally_check(cmd, value)
