import rx
import rx.disposable
import rx.operators as ops
from rx.core.abc import Disposable
from rx.core.typing import Observable

import sfm2._implementation.commands.command_list as cmds
from sfm2._implementation.commands.command import Command
from sfm2._implementation.commands.command_interface import CommandInterface
from sfm2._implementation.components.data_enabler import DataEnabler
from sfm2._implementation.internal_utils.sample_timestamper import SampleTimestamper
from sfm2._implementation.sample_parsing.binary_mode_frame import BinaryModeFrame, SampleTag
from sfm2.interface.sample_types import VectorSample, QuaternionSample, HeadingTiltSample, FloatSample, TimeSyncSample
from sfm2.interface.sfm2_data import Sfm2Data


class Sfm2DataImpl(Sfm2Data, Disposable):
    def __init__(self, cmd: CommandInterface, data: Observable[BinaryModeFrame]):
        self._enabler = DataEnabler(cmd)
        self._timestamper = SampleTimestamper()

        def create(enable: Command, sample_tag: SampleTag):
            return self._create_data_observable(enable, sample_tag, self._enabler, data)

        self._ad = create(cmds.ade, SampleTag.AD)
        self._gd = create(cmds.gde, SampleTag.GD)
        self._md = create(cmds.mde, SampleTag.MD)
        self._sfq = create(cmds.sfqde, SampleTag.SFQ)
        self._sfqt = create(cmds.sfqtde, SampleTag.SFQT)
        self._sfla = create(cmds.sflade, SampleTag.SFLA)
        self._sfea = create(cmds.sfeade, SampleTag.SFEA)
        self._sfcht = create(cmds.sfchtde, SampleTag.SFCHT)
        self._sfm = create(cmds.sfmde, SampleTag.SFM)
        self._pd = create(cmds.pde, SampleTag.PD)
        self._altd = create(cmds.altde, SampleTag.ALTD)
        self._td = create(cmds.tde, SampleTag.TD)
        self._hd = create(cmds.hde, SampleTag.HD)
        self._ts = create(cmds.tsde, SampleTag.TS)

        self._disposable = rx.disposable.Disposable()

    def start(self) -> None:
        self._disposable = self._ts.subscribe(self._timestamper.on_time_sync)

    @property
    def ad(self) -> Observable[VectorSample]:
        return self._ad

    @property
    def gd(self) -> Observable[VectorSample]:
        return self._gd

    @property
    def md(self) -> Observable[VectorSample]:
        return self._md

    @property
    def sfq(self) -> Observable[QuaternionSample]:
        return self._sfq

    @property
    def sfqt(self) -> Observable[QuaternionSample]:
        return self._sfqt

    @property
    def sfla(self) -> Observable[VectorSample]:
        return self._sfla

    @property
    def sfea(self) -> Observable[VectorSample]:
        return self._sfea

    @property
    def sfcht(self) -> Observable[HeadingTiltSample]:
        return self._sfcht

    @property
    def sfm(self) -> Observable[VectorSample]:
        return self._sfm

    @property
    def pd(self) -> Observable[FloatSample]:
        return self._pd

    @property
    def altd(self) -> Observable[FloatSample]:
        return self._altd

    @property
    def td(self) -> Observable[FloatSample]:
        return self._td

    @property
    def hd(self) -> Observable[FloatSample]:
        return self._hd

    @property
    def ts(self) -> Observable[TimeSyncSample]:
        return self._ts

    def reenable_streams(self) -> None:
        self._enabler.reenable()

    def dispose(self):
        self._disposable.dispose()

    def _create_data_observable(self, enable: Command, sample_tag: SampleTag, enabler: DataEnabler,
                                data: Observable[BinaryModeFrame]) -> Observable:
        def create_observable(_):
            return data.pipe(
                ops.map(lambda f: f.samples.get(sample_tag)),
                ops.filter(lambda s: s is not None),
                ops.map(self._timestamper.timestamp)
            )

        with_enable = rx.using(enabler.create_enable(enable),
                               create_observable)

        return with_enable.pipe(
            ops.publish(),
            ops.ref_count()
        )
