import logging
from typing import Union

import rx.operators as ops
from rx.core.typing import Observable

from sfm2.interface.env_sensor_config_types import PressureSampleRate, TemperatureSamplingSetup, TemperatureSamplingMode
from sfm2.interface.exceptions import InvalidStateException
from sfm2.interface.sample_types import FloatSample
from sfm2.interface.sfm2_config import SampleRate, Sfm2Config
from sfm2.interface.sfm2_data import Sfm2Data
from sfm2.interface.sfm2_env_sensors import Sfm2EnvSensors
from sfm2.interface.sfm2_init_options import Sfm2InitOptions

_log = logging.getLogger(__name__)


class Sfm2EnvSensorsImpl(Sfm2EnvSensors):
    def __init__(self, config: Sfm2Config, data: Sfm2Data, options: Sfm2InitOptions):
        self._config = config
        self._data = data
        self._options = options

    @property
    def pressure(self) -> Observable[FloatSample]:
        return self._data.pd

    @property
    def altitude(self) -> Observable[FloatSample]:
        return self._data.altd

    @property
    def temperature(self) -> Observable[FloatSample]:
        return self._data.td

    @property
    def humidity(self) -> Observable[FloatSample]:
        return self._data.hd

    @property
    def pressure_altitude_sample_rate(self) -> PressureSampleRate:
        return self._config.psr

    @pressure_altitude_sample_rate.setter
    def pressure_altitude_sample_rate(self, rate: Union[PressureSampleRate, int]) -> None:
        rate_not_int = PressureSampleRate(rate)
        if rate_not_int != PressureSampleRate.OFF:
            self._enable_acc_if_necessary()
        self._config.psr = rate_not_int

    @property
    def temperature_humidity_sampling_setup(self) -> TemperatureSamplingSetup:
        sampling_mode = self._config.tmode
        interval = self._config.tint
        return TemperatureSamplingSetup(sampling_mode, interval)

    @temperature_humidity_sampling_setup.setter
    def temperature_humidity_sampling_setup(self, setup: TemperatureSamplingSetup) -> None:
        if setup.mode != TemperatureSamplingMode.OFF:
            self._enable_acc_if_necessary()
        self._config.tmode = setup.mode
        self._config.tint = setup.sampling_interval

    def tare_altitude(self) -> None:
        if self.pressure_altitude_sample_rate == PressureSampleRate.OFF:
            InvalidStateException("Can't tare altitude when sampling disabled.").raise_and_log_warning(_log)

        pressure = self.pressure.pipe(ops.take(1)).run()
        self._config.altzero = pressure

    def _enable_acc_if_necessary(self) -> None:
        if self._options.enable_acc_for_env_sensors \
                and self._config.asr == SampleRate.OFF \
                and self._config.gsr == SampleRate.OFF:
            _log.info("IMU sensors disabled, enabling ACC for env. sensors' timestamping to work."
                      " You can disable this feature by clearing Sfm2InitOptions.enable_acc_for_env_sensors")
            self._config.asr = SampleRate.HZ_104
