import logging
from typing import Union

from rx.core.typing import Observable

from sfm2._implementation.internal_utils.find_sample_rate import find_sample_rate
from sfm2.interface.sample_types import QuaternionSample, VectorSample, HeadingTiltSample
from sfm2.interface.sfm2_config import SampleRate, Sfm2Config
from sfm2.interface.sfm2_data import Sfm2Data
from sfm2.interface.sfm2_sensor_fusion import Sfm2SensorFusion

_log = logging.getLogger(__name__)


class Sfm2SensorFusionImpl(Sfm2SensorFusion):
    def __init__(self, config: Sfm2Config, data: Sfm2Data):
        self._config = config
        self._data = data

    @property
    def quaternion(self) -> Observable[QuaternionSample]:
        return self._data.sfqt

    @property
    def untared_quaternion(self) -> Observable[QuaternionSample]:
        return self._data.sfq

    @property
    def linear_acceleration(self) -> Observable[VectorSample]:
        return self._data.sfla

    @property
    def euler_angles(self) -> Observable[VectorSample]:
        return self._data.sfea

    @property
    def heading_tilt(self) -> Observable[HeadingTiltSample]:
        return self._data.sfcht

    @property
    def calibrated_magnetometer(self) -> Observable[VectorSample]:
        return self._data.sfm

    @property
    def rate(self) -> SampleRate:
        return self._config.sfor

    @rate.setter
    def rate(self, rate: Union[SampleRate, float]) -> None:
        found_rate = find_sample_rate(rate, maximum=SampleRate.HZ_833)
        self._configure_sensors_rates_if_necessary(found_rate)
        self._config.sfor = found_rate

    @property
    def output_prescaler(self) -> int:
        return self._config.sfop

    @output_prescaler.setter
    def output_prescaler(self, prescaler: int) -> None:
        self._config.sfop = prescaler

    def _configure_sensors_rates_if_necessary(self, requested_sfor: SampleRate) -> None:
        if requested_sfor > self._config.asr:
            _log.info(f"Settings accelerometer sample rate to match requested Sensor Fusion rate")
            self._config.asr = requested_sfor

        if requested_sfor > self._config.gsr:
            _log.info(f"Settings gyroscope sample rate to match requested Sensor Fusion rate")
            self._config.gsr = requested_sfor

        appropriate_msr = min(requested_sfor, SampleRate.HZ_104)

        if appropriate_msr > self._config.msr:
            _log.info(f"Settings magnetometer sample rate to appropriate value")
            self._config.msr = appropriate_msr
