import logging
from datetime import timedelta
from typing import Union

import rx.operators as ops
from rx.core.typing import Observable

import sfm2._implementation.commands.command_list as cmds
from sfm2._implementation.commands.command_interface import CommandInterface
from sfm2._implementation.internal_utils.sfm2_rtc_time_conversion import sfm2_rtc_to_timedelta, \
    sfm2_rtc_from_timedelta
from sfm2._implementation.internal_utils.wrap_exception import wrap_exception
from sfm2.interface.exceptions import CommandFailedException, TimeNotValidException
from sfm2.interface.sfm2_time import Sfm2Time

_log = logging.getLogger(__name__)


class Sfm2TimeImpl(Sfm2Time):
    _REPORTS_INTERVAL_BASE = 1024

    def __init__(self, cmd: CommandInterface):
        self._cmd = cmd
        self._reports = self._cmd.observe_values(cmds.time).pipe(
            ops.map(self._parse_time),
            ops.publish(),
            ops.ref_count()
        )

    @wrap_exception(exc_to_catch=CommandFailedException, wrapping_exc=TimeNotValidException)
    def get(self) -> timedelta:
        time_str = self._cmd.query(cmds.time)
        return self._parse_time(time_str)

    @wrap_exception(exc_to_catch=CommandFailedException, wrapping_exc=TimeNotValidException)
    def set(self, ts: timedelta) -> None:
        time_int = sfm2_rtc_from_timedelta(ts)
        _log.info(f"Setting time {ts} ({time_int})")
        self._cmd.set_value(cmds.time, str(time_int))

    @wrap_exception(exc_to_catch=CommandFailedException, wrapping_exc=TimeNotValidException)
    def reset(self) -> None:
        _log.info(f"Resetting time")
        self._cmd.command(cmds.time)

    @wrap_exception(exc_to_catch=CommandFailedException, wrapping_exc=TimeNotValidException)
    def apply_offset(self, offset: Union[timedelta, int]) -> None:
        if isinstance(offset, timedelta):
            offset_ticks = sfm2_rtc_from_timedelta(offset)
            self._apply_offset_ticks(offset_ticks)
        else:
            self._apply_offset_ticks(offset)

    def _apply_offset_ticks(self, offset: int) -> None:
        _log.debug(f"Applying time offset {offset}")
        self._cmd.set_value(cmds.toffset, str(offset))

    @property
    def reports(self) -> Observable[timedelta]:
        return self._reports

    @property
    def reports_rate_hz(self) -> float:
        interval_str = self._cmd.query(cmds.tinterval)
        interval = int(interval_str)
        if interval == 0:
            return 0

        rate = self._REPORTS_INTERVAL_BASE / interval
        return rate

    @reports_rate_hz.setter
    def reports_rate_hz(self, rate: float) -> None:
        interval = self._REPORTS_INTERVAL_BASE / rate \
            if rate != 0 \
            else 0
        self._cmd.set_value(cmds.tinterval, str(interval))

    @property
    def trim(self) -> int:
        trim_str = self._cmd.query(cmds.ttrim)
        return int(trim_str)

    @trim.setter
    def trim(self, trim: int) -> None:
        _log.debug(f"Setting time trim: {trim}")
        self._cmd.set_value(cmds.ttrim, str(trim))

    def calibrate(self, time_passed_on_sfm2: timedelta, time_passed_on_pc: timedelta) -> None:
        _log.debug(f"Calibrating. SFM2 time: {time_passed_on_sfm2}, PC: {time_passed_on_pc}")

        ticks_passed_on_device = self._get_ticks_passed_on_sfm2(time_passed_on_sfm2)
        ticks_passed_on_pc = sfm2_rtc_from_timedelta(time_passed_on_pc)
        new_trim = ticks_passed_on_device / (ticks_passed_on_pc - ticks_passed_on_device)
        new_trim_int = int(round(new_trim))
        self.trim = new_trim_int

    def _get_ticks_passed_on_sfm2(self, time_passed_on_sfm2: timedelta):
        ticks_trimmed = sfm2_rtc_from_timedelta(time_passed_on_sfm2)
        current_trim = self.trim
        if current_trim == 0:
            return ticks_trimmed

        return ticks_trimmed * current_trim / (current_trim + 1)

    @staticmethod
    def _parse_time(time_cmd_value) -> timedelta:
        time_int = int(time_cmd_value)
        return sfm2_rtc_to_timedelta(time_int)
