import logging
import math
from datetime import timedelta
from typing import Optional, List, Any

from sfm2._implementation.internal_utils.sfm2_rtc_time_conversion import sfm2_rtc_to_seconds
from sfm2.interface.sample_types import TimeSyncSample, Sample

_log = logging.getLogger(__name__)


class SampleTimestamper:
    def __init__(self):
        self._interval_filter = _ImuClockIntervalFilter()
        self._last_ts: Optional[TimeSyncSample] = None

    def on_time_sync(self, ts: TimeSyncSample) -> None:
        self._interval_filter.add_time_sync(ts)
        self._last_ts = ts

    def timestamp(self, sample: Sample) -> Sample:
        timestamp_s = self._calc_timestamp(sample)
        if math.isnan(timestamp_s):
            return sample

        timestamp = timedelta(seconds=timestamp_s)
        return sample.add_timestamp(timestamp)

    def _calc_timestamp(self, sample: Sample) -> float:
        if self._last_ts is None \
                or sample.device_timestamp < self._last_ts.device_timestamp \
                or not self._interval_filter.ready():
            return math.nan

        diff_from_last_sync = sample.device_timestamp - self._last_ts.device_timestamp
        time_from_last_sync = diff_from_last_sync * self._interval_filter.get_imu_clock_interval_s()

        last_sync_time = sfm2_rtc_to_seconds(self._last_ts.clock)
        return last_sync_time + time_from_last_sync


class _ImuClockIntervalFilter:
    _MAX_BACKLOG_LENGTH = 10
    _MAX_RELATIVE_DIFF = 0.03

    def __init__(self):
        self._backlog = _FixedLengthBacklog(max_size=self._MAX_BACKLOG_LENGTH)
        self._config_idx = 0
        self._interval_s = math.nan

    def ready(self) -> bool:
        return not math.isnan(self._interval_s)

    def get_imu_clock_interval_s(self) -> float:
        return self._interval_s

    def add_time_sync(self, ts: TimeSyncSample) -> None:
        if ts.config_idx != self._config_idx:
            _log.debug("Config idx chagned. Clearing backlog")
            self._backlog.clear()

        if len(self._backlog) > 0:
            interval_since_first = self._calc_imu_clock_interval_s(ts, self._backlog.get_earliest())
            interval_since_last = self._calc_imu_clock_interval_s(ts, self._backlog.get_latest())

            if self._is_error_within_limit(interval_since_last, interval_since_first):
                self._interval_s = interval_since_first
            else:
                _log.warning("Calculated IMU clock interval diff is too large. Clearing backlog.")
                self._backlog.clear()

        self._backlog.add(ts)
        self._config_idx = ts.config_idx

    @classmethod
    def _is_error_within_limit(cls, value1: float, value2: float) -> bool:
        rel_error = (value1 - value2) / value2
        return math.fabs(rel_error) < cls._MAX_RELATIVE_DIFF

    @classmethod
    def _calc_imu_clock_interval_s(cls, t1: TimeSyncSample, t2: TimeSyncSample) -> float:
        clock_diff = t2.clock - t1.clock
        imu_diff = t2.device_timestamp - t1.device_timestamp

        time_diff_s = sfm2_rtc_to_seconds(clock_diff)
        return time_diff_s / imu_diff


class _FixedLengthBacklog:
    def __init__(self, max_size: int):
        self._backlog: List = list()
        self._max_size = max_size
        self._earliest_element_index = 0

    def add(self, element: Any) -> None:
        if len(self._backlog) < self._max_size:
            self._backlog.append(element)
        else:
            self._backlog[self._earliest_element_index] = element
            self._earliest_element_index = self._increment_index(self._earliest_element_index)

    def clear(self):
        self._backlog.clear()
        self._earliest_element_index = 0

    def get_earliest(self) -> Any:
        return self._backlog[self._earliest_element_index]

    def get_latest(self) -> Any:
        if len(self._backlog) != self._max_size:
            return self._backlog[-1]
        else:
            idx = self._increment_index(self._earliest_element_index)
            return self._backlog[idx]

    def _increment_index(self, index: int) -> int:
        return (index + 1) % (self._max_size - 1)

    def __len__(self):
        return len(self._backlog)
