import queue
from enum import Enum
from threading import Thread
from typing import Any, Tuple

from rx.core.abc import Disposable, Subject, Observer

from sfm2._implementation.internal_utils.disposed_token import DisposedToken


class _ElementType(Enum):
    VALUE = 0,
    ERROR = 1,
    COMPLETED = 2,
    DISPOSED = 3,
    TIMEOUT = 4


class ThreadRawSubject(Subject, Disposable):
    def __init__(self):
        super().__init__()
        self._disposed = DisposedToken()
        self._queue = queue.SimpleQueue()
        self._observer = _NoObserver()

        thread = Thread(target=self._run)
        thread.daemon = True
        thread.start()

    def on_next(self, value: Any) -> None:
        self._put_to_queue(_ElementType.VALUE, value)
        l = self._queue.qsize()

    def on_error(self, error: Any) -> None:
        self._put_to_queue(_ElementType.ERROR, error)

    def on_completed(self) -> None:
        self._put_to_queue(_ElementType.COMPLETED, None)

    def subscribe(self, observer=None, *, scheduler=None) -> Disposable:
        assert (observer is not None)
        assert (scheduler is None)
        assert (isinstance(self._observer, _NoObserver))

        self._observer = observer
        return self

    def dispose(self) -> None:
        self._observer = _NoObserver()
        self._disposed.dispose()
        self._put_to_queue(_ElementType.DISPOSED, None)

    def _run(self):
        while not self._disposed:
            try:
                el_type, el = self._get_from_queue()
                if el_type == _ElementType.VALUE:
                    self._observer.on_next(el)
                elif el_type == _ElementType.ERROR:
                    self._observer.on_error(el)
                    self.dispose()
                elif el_type == _ElementType.COMPLETED:
                    self._observer.on_completed()
                    self.dispose()
            except Exception as e:
                self._observer.on_error(e)
                self.dispose()
        self._observer.on_completed()

    def _get_from_queue(self) -> Tuple[_ElementType, Any]:
        try:
            return self._queue.get(block=True, timeout=5)
        except queue.Empty:
            return _ElementType.TIMEOUT, None

    def _put_to_queue(self, el_type: _ElementType, el: Any) -> None:
        element = el_type, el
        self._queue.put(element, block=False)


class _NoObserver(Observer):
    def on_next(self, value):
        pass

    def on_error(self, error):
        pass

    def on_completed(self):
        pass
