import typing
from threading import Lock

import rx.disposable
from rx.core.abc import Disposable, Observer
from rx.subject import Subject

from sfm2._implementation.internal_utils.thread_raw_subject import ThreadRawSubject


class ThreadSharedSubject(Subject, Disposable):
    def __init__(self):
        super().__init__()
        self._composite_observer = _CompositeObserver()
        self._raw_subject = ThreadRawSubject()
        self._subscription = self._raw_subject.subscribe(self._composite_observer)

    def on_next(self, value):
        self._raw_subject.on_next(value)

    def on_error(self, error):
        self._raw_subject.on_error(error)

    def on_completed(self):
        self._raw_subject.on_completed()

    def _subscribe_core(self,
                        observer: rx.typing.Observer,
                        scheduler: typing.Optional[rx.typing.Scheduler] = None
                        ) -> rx.typing.Disposable:
        return self._composite_observer.register(observer)

    def dispose(self):
        self._subscription.dispose()
        self._raw_subject.dispose()
        self._composite_observer.dispose()


class _CompositeObserver(Observer, Disposable):
    def __init__(self):
        self._lock = Lock()
        self._editable_observers: typing.Set[Observer] = set()
        self._published_observers: typing.Tuple[Observer, ...] = tuple()
        self._publish_observers()

    def register(self, observer: Observer) -> Disposable:
        with self._lock:
            self._editable_observers.add(observer)
            self._publish_observers()

        return rx.disposable.Disposable(action=lambda: self._unregister(observer))

    def on_next(self, value):
        self._for_each_observer(lambda o: o.on_next(value))

    def on_error(self, error):
        self._for_each_observer(lambda o: o.on_error(error))

    def on_completed(self):
        self._for_each_observer(lambda o: o.on_completed())

    def dispose(self):
        self.on_completed()

        with self._lock:
            self._editable_observers.clear()
            self._publish_observers()

    def _for_each_observer(self, action: typing.Callable[[Observer], None]) -> None:
        obs = self._published_observers
        for o in obs:
            action(o)

    def _unregister(self, observer: Observer) -> None:
        with self._lock:
            self._editable_observers.remove(observer)
            self._publish_observers()

    def _publish_observers(self) -> None:
        self._published_observers = tuple(self._editable_observers)
