from dataclasses import dataclass
from datetime import timedelta
from struct import Struct
from typing import Tuple, Optional, Callable, Iterable, Type

from sfm2._implementation.sample_parsing.binary_mode_frame import SampleTag
from sfm2.interface.sample_types import SampleType, SampleFieldType, Sample, QuaternionSample, \
    HeadingTiltSample, FloatSample, TimeSyncSample, VectorSample


@dataclass(frozen=True)
class SampleFactory:
    tag: SampleTag
    length_floats: int
    create: Callable[[int, bytes, int], Sample]


def create_sample_factories() -> Iterable[SampleFactory]:
    yield _vector(tag=SampleTag.AD, unit="g")
    yield _vector(tag=SampleTag.GD, unit="°/s")
    yield _vector(tag=SampleTag.MD, unit="µT")

    yield _quaternion(tag=SampleTag.SFQ)
    yield _quaternion(tag=SampleTag.SFQT)
    yield _vector(tag=SampleTag.SFLA, unit="g")
    yield _vector(tag=SampleTag.SFEA, unit="°")
    yield _heading_tilt(tag=SampleTag.SFCHT)
    yield _vector(tag=SampleTag.SFM, unit="µT")

    yield _single_value(tag=SampleTag.PD, field_name="pressure", unit="hPa")
    yield _single_value(tag=SampleTag.ALTD, field_name="altitude", unit="m")
    yield _single_value(tag=SampleTag.TD, field_name="temperature", unit="°C")
    yield _single_value(tag=SampleTag.HD, field_name="humidity", unit="%")

    yield _time_sync(tag=SampleTag.TS)


def _vector(tag: SampleTag, unit: str) -> SampleFactory:
    sample_type = SampleType(name=tag.value, fields=_same_unit_fields(unit, "x", "y", "z"))
    return _create_factory_from_floats(tag, sample_type, VectorSample)


def _quaternion(tag: SampleTag) -> SampleFactory:
    sample_type = SampleType(name=tag.value, fields=_same_unit_fields(None, "w", "x", "y", "z"))
    return _create_factory_from_floats(tag, sample_type, QuaternionSample)


def _heading_tilt(tag: SampleTag) -> SampleFactory:
    sample_type = SampleType(name=tag.value, fields=_same_unit_fields("°", "heading", "tilt"))
    return _create_factory_from_floats(tag, sample_type, HeadingTiltSample)


def _single_value(tag: SampleTag, field_name: str, unit: str) -> SampleFactory:
    sample_type = SampleType(name=tag.value, fields=_same_unit_fields(unit, field_name))
    return _create_factory_from_floats(tag, sample_type, FloatSample)


def _time_sync(tag: SampleTag) -> SampleFactory:
    sample_type = SampleType(name=tag.value, fields=_same_unit_fields(None, "Time", "Config Idx"))

    def _create(ts: int, data: bytes, offset: int) -> TimeSyncSample:
        fields = Struct(f"<II").unpack_from(data, offset)
        return TimeSyncSample(type=sample_type, timestamp=None, device_timestamp=ts, fields=fields)

    return SampleFactory(tag=tag, length_floats=2, create=_create)


def _create_factory_from_floats(tag: SampleTag, sample_type: SampleType, sample_object_type: Type) -> SampleFactory:
    length_floats = len(sample_type.fields)

    def _create(ts: int, data: bytes, offset: int) -> Sample:
        floats = Struct(f"<{length_floats}f").unpack_from(data, offset)
        return sample_object_type(type=sample_type, timestamp=None, device_timestamp=ts, fields=tuple(floats[:length_floats]))

    return SampleFactory(tag=tag, length_floats=length_floats, create=_create)


def _same_unit_fields(unit: Optional[str], *names: str) -> Tuple[SampleFieldType, ...]:
    return tuple(SampleFieldType(name=n, unit=unit) for n in names)


def _vector_fields(unit: Optional[str]) -> Tuple[SampleFieldType, ...]:
    return _same_unit_fields(unit, "x", "y", "z")
