from struct import Struct
from typing import Iterable

from sfm2._implementation.internal_utils.sfm2_rtc_time_conversion import sfm2_rtc_to_timedelta
from sfm2._implementation.sample_parsing.binary_mode_frame import BinaryModeFrame
from sfm2._implementation.sample_parsing.sample_factories import SampleFactory, create_sample_factories
from sfm2.interface.exceptions import MalformedDataException

_HEADER_SIZE = 3
_TS_SIZE = 4
_FOOTER_SIZE = 1
_START_BYTE = 0xFA
_END_BYTE = 0xFB


class BinaryFrameParser:
    def __init__(self):
        self._factories = list(create_sample_factories())

    @property
    def start_byte(self) -> int:
        return _START_BYTE

    @property
    def end_byte(self) -> int:
        return _END_BYTE

    @property
    def header_size(self) -> int:
        return _HEADER_SIZE

    def get_frame_size(self, frame_front: bytes) -> int:
        sample_factories = self._get_sample_factories(frame_front)
        length_floats = sum(f.length_floats for f in sample_factories)

        return _HEADER_SIZE + _TS_SIZE + length_floats * 4 + _FOOTER_SIZE

    def parse(self, frame: bytes) -> BinaryModeFrame:
        samples = dict()
        offset = self.header_size

        ts, = Struct("<I").unpack_from(frame, offset)
        offset += _TS_SIZE

        for f in self._get_sample_factories(frame):
            samples[f.tag] = f.create(ts, frame, offset)
            offset += f.length_floats * 4

        return BinaryModeFrame(samples)

    def _get_sample_factories(self, frame_front: bytes) -> Iterable[SampleFactory]:
        header = frame_front[1] + 0x100 * frame_front[2]
        if header >= (1 << len(self._factories)):
            MalformedDataException("Invalid frame header").raise_and_log_error()

        for idx, f in enumerate(self._factories):
            if header & (1 << idx) != 0:
                yield f
