from abc import ABC, abstractmethod
from typing import Callable, Optional

from sfm2._implementation.sample_parsing.sample_parser import BinaryModeFrame, BinaryFrameParser
from sfm2._implementation.serial_port.bytes_stream import BytesStream
from sfm2.interface.exceptions import MalformedDataException


class Packetizer(ABC):
    @abstractmethod
    def process(self, data: BytesStream) -> bool:
        pass

    def process_all(self, data: BytesStream) -> None:
        while data.unexamined():
            self.process(data)


class TextPacketizer(Packetizer):
    _NEWLINE = ord('\n')

    def __init__(self, on_packet: Callable[[str], None]):
        self._on_packet = on_packet

    def process(self, data: BytesStream) -> bool:
        for b in data:
            if b == self._NEWLINE:
                packet = data.consume_examined()
                line = packet[:-2].decode("ascii")
                self._on_packet(line)
                return True
        return False


class BinaryPacketizer(Packetizer):
    _SIZE_UNKNOWN = 0

    def __init__(self, parser: BinaryFrameParser, on_packet: Callable[[BinaryModeFrame], None]):
        self._parser = parser
        self._on_packet = on_packet
        self._frame_size = self._SIZE_UNKNOWN

    def process(self, data: BytesStream) -> bool:
        size_known = self._frame_size != self._SIZE_UNKNOWN
        if size_known and len(data) >= self._frame_size:
            frame_bytes = data.consume(self._frame_size)
            self._frame_size = self._SIZE_UNKNOWN
            if frame_bytes[-1] != self._parser.end_byte:
                raise MalformedDataException("Invalid packet end")
            frame = self._parser.parse(frame_bytes)
            self._on_packet(frame)
            return True

        elif not size_known and len(data) >= self._parser.header_size:
            self._frame_size = self._parser.get_frame_size(data.content)

        else:
            data.all_examined()

        return False


class CompositePacketizer(Packetizer):
    def __init__(self, binary_parser: BinaryFrameParser, on_text: Callable[[str], None],
                 on_binary: Callable[[BinaryModeFrame], None]):
        self._binary_frame_start_byte = binary_parser.start_byte
        self._text_packetizer = TextPacketizer(on_text)
        self._binary_packetizer = BinaryPacketizer(binary_parser, on_binary)
        self._current_packetizer: Optional[Packetizer] = None

    def process(self, data: BytesStream) -> bool:
        if self._current_packetizer is not None:
            completed = self._current_packetizer.process(data)
            if completed:
                self._current_packetizer = None
            return completed
        elif len(data) >= 1:
            self._current_packetizer = self._choose_packetizer(data)

        return False

    def _choose_packetizer(self, data: BytesStream) -> Packetizer:
        first_byte = next(data)
        if first_byte == self._binary_frame_start_byte:
            return self._binary_packetizer
        elif first_byte < 128:
            return self._text_packetizer
        else:
            data.consume(len(data))
            raise MalformedDataException(f"Unexpected frame start: {first_byte}")
