import logging
from threading import Thread
from time import sleep

import rx
import rx.operators as ops
import serial
from rx.core.abc import Disposable
from rx.core.typing import Observable
from rx.subject import BehaviorSubject

from sfm2._implementation.internal_utils.thread_shared_subject import ThreadSharedSubject
from sfm2._implementation.sample_parsing.binary_mode_frame import BinaryModeFrame
from sfm2._implementation.sample_parsing.sample_parser import BinaryFrameParser
from sfm2._implementation.serial_port.bytes_stream import BytesStream
from sfm2._implementation.serial_port.packetizer import CompositePacketizer
from sfm2.interface.exceptions import SerialPortNotActiveException, InvalidStateException, SerialPortRxTimeoutException, \
    SerialPortSynchronizationFailed
from sfm2.interface.sfm2_init_options import Sfm2InitOptions

_log = logging.getLogger(__name__)

_BAUDRATE = 921600

_PING_REQUEST = "PING!"
_PING_RESPONSE = b'PING=OK\r\n'
_PINGING_INTERVAL_S = 0.3


class SerialPort(Disposable):
    def __init__(self, port_name: str, options: Sfm2InitOptions):
        self._options: Sfm2InitOptions = options
        self._binary_frame_parser = BinaryFrameParser()

        self._running = False
        self._synchronized_subject = BehaviorSubject(value=False)
        self._txt_subject = ThreadSharedSubject()
        self._bin_subject = ThreadSharedSubject()

        if options.log_all_cmds:
            self._txt_subject.subscribe(lambda l: _log.debug(f"Received: {l}"))

        _log.debug(f"Opening serial port {port_name}")
        self._port = serial.Serial(port_name, baudrate=_BAUDRATE, rtscts=True)
        self._port.timeout = options.serial_rx_timeout_s
        self._configure_port_buffers()
        _log.info(f"Opened serial port {port_name}")

    def send(self, line: str) -> None:
        if self._options.log_all_cmds:
            _log.debug(f"Sending: {line}")

        if not self._running:
            SerialPortNotActiveException().raise_and_log_info(_log)

        as_bytes = f"{line}\r\n".encode('ascii')
        self._port.write(as_bytes)
        self._port.flush()

    @property
    def incoming_text(self) -> Observable[str]:
        return self._txt_subject

    @property
    def incoming_binary_mode_frames(self) -> Observable[BinaryModeFrame]:
        return self._bin_subject

    def start_listening(self) -> None:
        _log.debug("Starting listening")
        if self._running:
            InvalidStateException("Serial port already listening").raise_and_log_error(_log)

        thread = Thread(target=self._reading_loop)
        thread.daemon = True
        self._running = True
        thread.start()

    def wait_until_synchronized(self):
        try:
            self._synchronized_subject.pipe(
                ops.first(lambda s: s),
                ops.timeout(self._options.serial_port_sync_timeout_s)
            ).run()

        except Exception as e:
            SerialPortSynchronizationFailed().raise_and_log_fatal(log=_log, cause=e)

    def dispose(self):
        _log.info("Stopping listening and closing serial port")
        self._running = False
        self._port.close()
        self._txt_subject.dispose()
        self._bin_subject.dispose()
        self._synchronized_subject.dispose()

    def _close_port(self):
        self._port.rtscts = False
        self._port.setRTS(0)
        self._port.close()

    def _reading_loop(self) -> None:
        while self._running:
            try:
                self._sync()
                self._try_reading()
            except serial.SerialException as e:
                _log.error("Serial error encountered - can't recover")
                self._handle_fatal_error(e)
                break
            except Exception as e:
                if self._options.terminate_serial_on_error:
                    self._handle_fatal_error(e)
                    break
                elif self._running:
                    self._synchronized_subject.on_next(False)
                    _log.warning("Error encountered, will try to recover", exc_info=e)
                    sleep(1)

        self._txt_subject.on_completed()
        self._bin_subject.on_completed()

    def _try_reading(self) -> None:
        stream = BytesStream()
        packetizer = CompositePacketizer(
            binary_parser=self._binary_frame_parser,
            on_text=self._txt_subject.on_next,
            on_binary=self._bin_subject.on_next
        )

        while self._running:
            data = self._read_from_port(with_timeout=stream.any())
            stream.append(data)
            packetizer.process_all(stream)

    def _read_from_port(self, with_timeout: bool) -> bytes:
        bytes_to_read = self._port.inWaiting()
        if bytes_to_read == 0:
            bytes_to_read = 1

        data = self._port.read(bytes_to_read)
        if with_timeout and len(data) == 0:
            SerialPortRxTimeoutException().raise_and_log_warning()
        return data

    def _handle_fatal_error(self, e: Exception) -> None:
        if self._running:
            _log.exception("Fatal error encountered, terminating")
            self._txt_subject.on_error(e)
            self._bin_subject.on_error(e)
        else:
            _log.debug("Encountered error when already disposed, exiting gracefully")

    def _configure_port_buffers(self):
        # noinspection PyBroadException
        try:
            self._port.set_buffer_size(rx_size=128 * 1024, tx_size=4 * 1024)
        except Exception:
            _log.debug("Can't configure serial port buffer size. It works only on Windows")

    def _sync(self) -> None:
        if not self._options.serial_port_sync_enabled:
            self._synchronized_subject.on_next(True)
            _log.info("Serial port sync disabled")
            return

        _log.debug("Synchronizing serial port")

        while self._running:
            try:
                self._try_synchronizing()
                self._synchronized_subject.on_next(True)
                break
            except TimeoutError:
                _log.info("Nothing received from serial port. Trying reopening.")
                self._port.close()
                sleep(0.5)
                self._port.open()

        _log.debug("Stopping synchronization")

    def _try_synchronizing(self) -> None:
        self._port.reset_input_buffer()

        with rx.interval(_PINGING_INTERVAL_S).subscribe(lambda _: self.send(_PING_REQUEST)):
            while self._running:
                ok = self._try_reading_ping_response()
                if ok:
                    return

    def _try_reading_ping_response(self) -> bool:
        for byte_expected_int in _PING_RESPONSE:
            byte_expected = bytes([byte_expected_int])

            byte_read = self._port.read(1)

            if len(byte_read) == 0:
                raise TimeoutError

            if byte_read != byte_expected:
                return False
        return True
