import logging
from typing import Iterator

import serial.tools.list_ports

from sfm2._implementation.internal_utils.dispose_on_exception import DisposeOnException
from sfm2._implementation.internal_utils.wrap_exception import wrap_exception
from sfm2._implementation.sfm2_on_serial_port import Sfm2OnSerialPort
from sfm2.interface.exceptions import NoSfm2FoundException
from sfm2.interface.found_sfm2 import FoundSfm2
from sfm2.interface.sfm2 import Sfm2
from sfm2.interface.sfm2_init_options import Sfm2InitOptions, Sfm2CollectionInitOptions
from sfm2.utils.sfm2_collection import Sfm2Collection

_log = logging.getLogger(__name__)

_USB_VID = 0x10c4
_USB_PID = 0xea60


@wrap_exception(exc_to_catch=StopIteration, wrapping_exc=NoSfm2FoundException)
def connect_to_first_found_sfm2(options: Sfm2InitOptions) -> Sfm2:
    all_found = discover_connected_sfm2s(options)
    first = next(all_found)
    return first.connect()


def connect_to_all_found_sfm2s(device_options: Sfm2InitOptions,
                               collection_options: Sfm2CollectionInitOptions) -> Sfm2Collection:
    with DisposeOnException() as disposables:
        connected = list()
        for found in discover_connected_sfm2s(device_options):
            try:
                _log.debug(f"Connecting to {found.label}")
                c = found.connect()
                connected.append(c)
                disposables.add(c)

            except Exception:
                ignore = collection_options.ignore_connection_errors
                _log.exception(f"Could not connect to {found.label}. Ignoring: {ignore}")
                if not ignore:
                    raise
        return Sfm2Collection(connected, collection_options)


def discover_connected_sfm2s(options: Sfm2InitOptions) -> Iterator[FoundSfm2]:
    _log.info("Starting SFM2 discovery")
    all_available_ports = serial.tools.list_ports.comports()
    for port in sorted(all_available_ports):
        if port.pid == _USB_PID and port.vid == _USB_VID:
            _log.info(f"Found SFM2 on port {port.device}")
            yield _SerialPortWithSfm2(port.device, options)


class _SerialPortWithSfm2(FoundSfm2):
    def __init__(self, port: str, options: Sfm2InitOptions):
        self._port = port
        self._options = options

    @property
    def label(self) -> str:
        return self._port

    def connect(self) -> Sfm2:
        return Sfm2OnSerialPort(self._port, self._options)
