import logging
from threading import RLock
from time import sleep
from typing import Tuple

import sfm2._implementation.commands.command_list as cmds
from sfm2._implementation.commands.command_interface import CommandInterface
from sfm2._implementation.components.self_test import run_self_test
from sfm2._implementation.components.sfm2_accelerometer_impl import Sfm2AccelerometerImpl
from sfm2._implementation.components.sfm2_config_impl import Sfm2ConfigImpl
from sfm2._implementation.components.sfm2_data_impl import Sfm2DataImpl
from sfm2._implementation.components.sfm2_env_sensors_impl import Sfm2EnvSensorsImpl
from sfm2._implementation.components.sfm2_gyroscope_impl import Sfm2GyroscopeImpl
from sfm2._implementation.components.sfm2_magnetometer_impl import Sfm2MagnetometerImpl
from sfm2._implementation.components.sfm2_name import Sfm2Name
from sfm2._implementation.components.sfm2_reference_impl import Sfm2ReferenceImpl
from sfm2._implementation.components.sfm2_sensor_fusion_impl import Sfm2SensorFusionImpl
from sfm2._implementation.components.sfm2_time_impl import Sfm2TimeImpl
from sfm2._implementation.components.version_check import check_firmware_version
from sfm2._implementation.internal_utils.dispose_on_exception import DisposeOnException
from sfm2._implementation.serial_port.serial_port import SerialPort
from sfm2.interface.exceptions import CommandResponseTimeoutException, Sfm2Exception
from sfm2.interface.sfm2 import Sfm2
from sfm2.interface.sfm2_accelerometer import Sfm2Accelerometer
from sfm2.interface.sfm2_config import Sfm2Config
from sfm2.interface.sfm2_data import Sfm2Data
from sfm2.interface.sfm2_env_sensors import Sfm2EnvSensors
from sfm2.interface.sfm2_gyroscope import Sfm2Gyroscope
from sfm2.interface.sfm2_init_options import Sfm2InitOptions
from sfm2.interface.sfm2_magnetometer import Sfm2Magnetometer
from sfm2.interface.sfm2_reference import Sfm2Reference
from sfm2.interface.sfm2_sensor_fusion import Sfm2SensorFusion
from sfm2.interface.sfm2_time import Sfm2Time

_log = logging.getLogger(__name__)

_API_VERSION_RANGE: Tuple[int, int] = (4, 4)


class Sfm2OnSerialPort(Sfm2):
    def __init__(self, port_name: str, options: Sfm2InitOptions):
        _log.info(f"Connecting to SFM2 on port {port_name}")
        _log.debug(f"Init options: {options}")
        with DisposeOnException() as disposables:
            lock = RLock()
            self._port_name = port_name
            self._options = options
            self._port = SerialPort(port_name, options)
            disposables.add(self._port)
            self._cmd = CommandInterface(self._port, lock, options=options)
            self._name = Sfm2Name(self._cmd, options)
            self._time = Sfm2TimeImpl(self._cmd)
            self._config = Sfm2ConfigImpl(self._cmd, options)
            self._data = Sfm2DataImpl(self._cmd, self._port.incoming_binary_mode_frames)
            self._reference = Sfm2ReferenceImpl(self._cmd, self._config)

            self._sensor_fusion = Sfm2SensorFusionImpl(self._config, self._data)
            self._accelerometer = Sfm2AccelerometerImpl(self._config, self._data)
            self._gyroscope = Sfm2GyroscopeImpl(self._config, self._data)
            self._magnetometer = Sfm2MagnetometerImpl(self._config, self._data)
            self._env_sensors = Sfm2EnvSensorsImpl(self._config, self._data, options)

            self._port.start_listening()
            self._port.wait_until_synchronized()
            _log.debug("Serial port synchronized")

            self._check_firmware_version(options)
            self._data.start()

    @property
    def id(self) -> str:
        return self._port_name

    @property
    def name(self) -> str:
        return self._name.get()

    @name.setter
    def name(self, value: str) -> None:
        self._name.set(value)

    @property
    def sensor_fusion(self) -> Sfm2SensorFusion:
        return self._sensor_fusion

    @property
    def accelerometer(self) -> Sfm2Accelerometer:
        return self._accelerometer

    @property
    def gyroscope(self) -> Sfm2Gyroscope:
        return self._gyroscope

    @property
    def magnetometer(self) -> Sfm2Magnetometer:
        return self._magnetometer

    @property
    def env_sensors(self) -> Sfm2EnvSensors:
        return self._env_sensors

    @property
    def time(self) -> Sfm2Time:
        return self._time

    @property
    def reference(self) -> Sfm2Reference:
        return self._reference

    @property
    def config(self) -> Sfm2Config:
        return self._config

    @property
    def data(self) -> Sfm2Data:
        return self._data

    @property
    def version(self) -> str:
        return self._cmd.query(cmds.fwver)

    def reset(self) -> None:
        self._send_reset_with_retries()
        sleep(self._options.delay_after_reset_s)
        self._data.reenable_streams()

    def run_self_test(self) -> bool:
        return run_self_test(self._cmd, self._options)

    def __str__(self):
        return f"SFM2 on port {self._port_name}"

    def dispose(self) -> None:
        try:
            self._data.dispose()
            _log.debug("Disposing serial port")
            self._port.dispose()
        except Exception:
            _log.debug("Could not dispose serial port", exc_info=True)

    def _check_firmware_version(self, options: Sfm2InitOptions):
        check_firmware_version(self._cmd, _API_VERSION_RANGE, options.skip_version_check)

    def _send_reset_with_retries(self):
        trials_count = self._options.reset_retries_after_timeout + 1
        for trial in range(trials_count):
            try:
                self._cmd.command_and_check_if_ok(cmds.sfreset)
                return
            except CommandResponseTimeoutException:
                _log.info(f"Reset timeout, trial {trial + 1} of {trials_count}")

        Sfm2Exception(f"Reset failed even after {self._options.reset_retries_after_timeout} retries."
                      f" Make sure correct serial port is selected. Try turning the device OFF and ON again") \
            .raise_and_log_error(log=_log)
