import logging
from typing import Optional

_log = logging.getLogger(__name__)


class Sfm2Exception(Exception):
    def raise_and_log_fatal(self, log=_log, cause: Optional[Exception] = None):
        self._raise_and_log(level=logging.FATAL, log=log, cause=cause)

    def raise_and_log_error(self, log=_log, cause: Optional[Exception] = None):
        self._raise_and_log(level=logging.ERROR, log=log, cause=cause)

    def raise_and_log_warning(self, log=_log, cause: Optional[Exception] = None):
        self._raise_and_log(level=logging.WARNING, log=log, cause=cause)

    def raise_and_log_info(self, log=_log, cause: Optional[Exception] = None):
        self._raise_and_log(level=logging.INFO, log=log, cause=cause)

    def _raise_and_log(self, level: int, log, cause: Optional[Exception]):
        log.log(level=level, msg=f"Raising {type(self).__name__}: {self}", exc_info=cause)
        if cause is not None:
            raise self from cause
        else:
            raise self


class Sfm2InternalException(Sfm2Exception):
    def __init__(self, msg: str):
        super().__init__(f"Internal library error: {msg}")


class NoSfm2FoundException(Sfm2Exception):
    def __init__(self):
        super().__init__(f"No SFM2s found")


class SerialPortSynchronizationFailed(Sfm2Exception):
    def __init__(self):
        super().__init__("Serial port synchronization timeout.\n"
                         "You can try increasing the synchronization timeout using Sfm2InitOptions.serial_port_sync_timeout_s or"
                         " disable the synchronization using Sfm2InitOptions.serial_port_sync_enabled")


class MalformedDataException(Sfm2Exception):
    def __init__(self, msg: str):
        super().__init__(f"Malformed data received: {msg}")


class UnsupportedFirmwareVersionException(Sfm2Exception):
    def __init__(self, msg: str):
        super().__init__(f"Unsupported device firmware version. {msg}\n"
                         " You can disable firmware version check by setting Sfm2InitOptions.skip_version_check.")


class InvalidStateException(Sfm2Exception):
    pass


class TimeNotValidException(InvalidStateException):
    def __init__(self):
        super().__init__("Time is not valid when IMU not sampling."
                         " Set any sensors sampling rate and wait for at least one sample to be collected.")


class SerialPortNotActiveException(InvalidStateException):
    def __init__(self):
        super().__init__("Serial port not active. Check if Sfm2 was not disposed.")


class CommandResponseTimeoutException(Sfm2Exception):
    def __init__(self, command_name: str, timeout_s: float):
        super().__init__(f"Did not receive response to command {command_name} in {timeout_s}s")


class UnexpectedDataException(Sfm2Exception):
    pass


class CommandResponseValueMismatch(UnexpectedDataException):
    def __init__(self, command_name: str, set_value: str, got_value: str):
        super().__init__(f"Set {command_name} to {set_value} but got {got_value}."
                         " You can disable checking response values for some commands"
                         " by clearing Sfm2InitOptions.verify_all_cmd_response_values.")


class CommandFailedException(UnexpectedDataException):
    def __init__(self, command_name: str):
        super().__init__(f"Command {command_name} failed. Device responded with failure.")


class SerialPortRxTimeoutException(Sfm2Exception):
    def __init__(self):
        super().__init__(f"Serial RX timeouted.")


class InvalidArgumentException(Sfm2Exception):
    pass


class DuplicateKeyException(Sfm2Exception):
    pass
