from __future__ import annotations

import math
from dataclasses import dataclass
from datetime import timedelta
from typing import Tuple, Optional


@dataclass(frozen=True)
class SampleFieldType:
    """
    Describes the type of a single field in Sample
    """
    name: str
    unit: Optional[str]


@dataclass(frozen=True)
class SampleType:
    """
    Describes the type of a Sample
    """
    name: str
    fields: Tuple[SampleFieldType, ...]


@dataclass(frozen=True)
class Sample:
    """
    Base class for data sample contains a collection of values (fields) as well as the definition of the sample type.
    """
    type: SampleType
    timestamp: Optional[timedelta]
    device_timestamp: int
    fields: Tuple

    def __str__(self):
        def format_field_value(value: float, field: SampleFieldType):
            unit = field.unit if field.unit is not None else ""
            return f"{field.name}={value:0.3f}{unit}"

        field_strs = (format_field_value(v, f) for v, f in zip(self.fields, self.type.fields))
        joined = ", ".join(field_strs)

        timestamp_str = f" @{self.timestamp}" if self.timestamp is not None else ""

        return f"{self.type.name}: {joined}{timestamp_str}"

    def add_timestamp(self, timestamp: timedelta) -> Sample:
        # noinspection PyArgumentList
        return self.__class__(
            type=self.type,
            timestamp=timestamp,
            device_timestamp=self.device_timestamp,
            fields=self.fields
        )


class VectorSample(Sample):
    """
    Data sample which value is a 3-dimensional vector
    """

    @property
    def x(self) -> float:
        return self.fields[0]

    @property
    def y(self) -> float:
        return self.fields[1]

    @property
    def z(self) -> float:
        return self.fields[2]

    @property
    def magnitude(self) -> float:
        x = self.x
        y = self.y
        z = self.z
        return math.sqrt(x * x + y * y + z * z)


class QuaternionSample(Sample):
    """
    Data sample which value is a quaternion
    """

    @property
    def w(self) -> float:
        return self.fields[0]

    @property
    def x(self) -> float:
        return self.fields[1]

    @property
    def y(self) -> float:
        return self.fields[2]

    @property
    def z(self) -> float:
        return self.fields[3]


class HeadingTiltSample(Sample):
    """
    Data sample which value consists of two angles: heading and tilt.
    """

    @property
    def heading(self) -> float:
        return self.fields[0]

    @property
    def tilt(self) -> float:
        return self.fields[1]


class FloatSample(float, Sample):
    """
    Data sample which value is single float value. Can be used directly in place of a float value.
    """

    # noinspection PyShadowingBuiltins
    def __new__(cls, type: SampleType, timestamp: Optional[timedelta], device_timestamp: int,
                fields: Tuple[float, ...]):
        # noinspection PyArgumentList
        return float.__new__(cls, fields[0])

    # noinspection PyShadowingBuiltins
    def __init__(self, type: SampleType, timestamp: Optional[timedelta], device_timestamp: int,
                 fields: Tuple[float, ...]):
        float.__init__(fields[0])
        # noinspection PyArgumentList
        Sample.__init__(self, type, timestamp, device_timestamp, fields)

    @property
    def value(self) -> float:
        return self.fields[0]


class TimeSyncSample(Sample):
    """
    Time sync signal matching device timestamp with a precision clock.
    """

    @property
    def clock(self) -> int:
        return self.fields[0]

    @property
    def config_idx(self) -> int:
        return self.fields[1]
