from abc import abstractmethod

from rx.core.abc import Disposable

from sfm2.interface.sfm2_accelerometer import Sfm2Accelerometer
from sfm2.interface.sfm2_config import Sfm2Config
from sfm2.interface.sfm2_data import Sfm2Data
from sfm2.interface.sfm2_env_sensors import Sfm2EnvSensors
from sfm2.interface.sfm2_gyroscope import Sfm2Gyroscope
from sfm2.interface.sfm2_magnetometer import Sfm2Magnetometer
from sfm2.interface.sfm2_reference import Sfm2Reference
from sfm2.interface.sfm2_sensor_fusion import Sfm2SensorFusion
from sfm2.interface.sfm2_time import Sfm2Time


class Sfm2(Disposable):
    """
    Main interface for a connected SFM2 device.

    The object can be disposed to disconnect from the device, close serial port and free resources.
    """

    @property
    @abstractmethod
    def id(self) -> str:
        """
        Gets the SFM2 ID

        Currently, the ID is always the Serial Port name

        :type: str
        """
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Gets or sets the SFM2s name.

        All characters must be ASCII and the maximum length is 16.

        The name is stored in persistent memory.

        :type: str
        :raises InvalidArgumentException: The name that was set does not meet requirements.
        """
        pass

    @name.setter
    @abstractmethod
    def name(self, value: str) -> None:
        pass

    @property
    @abstractmethod
    def sensor_fusion(self) -> Sfm2SensorFusion:
        """
        Gets the high-level interface for Sensor Fusion

        :rtype: Sfm2SensorFusion
        """
        pass

    @property
    @abstractmethod
    def accelerometer(self) -> Sfm2Accelerometer:
        """
        Gets the high-level interface for accelerometer

        :rtype: Sfm2Accelerometer
        """
        pass

    @property
    @abstractmethod
    def gyroscope(self) -> Sfm2Gyroscope:
        """
        Gets the high-level interface for gyroscope

        :rtype: Sfm2Gyroscope
        """
        pass

    @property
    @abstractmethod
    def magnetometer(self) -> Sfm2Magnetometer:
        """
        Gets the high-level interface for magnetometer

        :rtype: Sfm2Magnetometer
        """
        pass

    @property
    @abstractmethod
    def env_sensors(self) -> Sfm2EnvSensors:
        """
        Gets the high-level interface for environmental sensors:
            - Temperature
            - Humidity
            - Pressure
            - Altitude

        :rtype: Sfm2EnvSensors
        """
        pass

    @property
    @abstractmethod
    def reference(self) -> Sfm2Reference:
        """
        Gets the interface for reference frame manipulation.

        :rtype: Sfm2Reference
        """
        pass

    @property
    @abstractmethod
    def time(self) -> Sfm2Time:
        """
        Gets the interface for timestamp reading and manipulation

        :rtype: Sfm2Time
        """
        pass

    @property
    @abstractmethod
    def config(self) -> Sfm2Config:
        """
        Gets the low-level interface for config.

        :rtype: Sfm2Config
        """
        pass

    @property
    @abstractmethod
    def data(self) -> Sfm2Data:
        """
        Gets the low-level interface for reading sensor data.

        :rtype: Sfm2Config
        """
        pass

    @property
    @abstractmethod
    def version(self) -> str:
        """
        Reads the firmware version

        :rtype: str
        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """
        Resets the configuration to default.
        This disables all sensors.
        """
        pass

    @abstractmethod
    def run_self_test(self) -> bool:
        """ Runs the built-in self-test.

        It checks all IMU sensors. It may take a few seconds.
        Remember to keep the SFM2 still during the whole process!

        :rtype: bool
        :returns: True if the test passed, False if failed.
        """
        pass

    @property
    def sf(self) -> Sfm2SensorFusion:
        """
        Gets the high-level interface for Sensor Fusion. Alias for :func:`~sfm2.Sfm2.sensor_fusion`

        :rtype: Sfm2SensorFusion
        """
        return self.sensor_fusion

    @property
    def acc(self) -> Sfm2Accelerometer:
        """Gets the high-level interface for accelerometer.
        Alias for :py:attr:`~sfm2.interface.sfm2.Sfm2.accelerometer`

        :rtype: Sfm2Accelerometer
        """
        return self.accelerometer

    @property
    def gyr(self) -> Sfm2Gyroscope:
        """Gets the high-level interface for gyroscope.
         Alias for :py:attr:`~sfm2.interface.sfm2.Sfm2.gyroscope`

        :rtype: Sfm2Gyroscope
        """
        return self.gyroscope

    @property
    def mag(self) -> Sfm2Magnetometer:
        """Gets the high-level interface for magnetometer.
         Alias for :py:attr:`~sfm2.interface.sfm2.Sfm2.magnetometer`

        :rtype: Sfm2Magnetometer
        """
        return self.magnetometer

    @property
    def env(self) -> Sfm2EnvSensors:
        """
        Gets the high-level interface for environmental sensors:
            - Temperature
            - Humidity
            - Pressure
            - Altitude

        Alias for :py:attr:`~sfm2.interface.sfm2.Sfm2.env_sensors`

        :rtype: Sfm2EnvSensors
        """
        return self.env_sensors

    @property
    def ref(self) -> Sfm2Reference:
        """
        Gets the interface for reference frame manipulation.

        Alias for :py:attr:`~sfm2.interface.sfm2.Sfm2.reference`

        :rtype: Sfm2Reference
        """
        return self.reference
