from abc import ABC, abstractmethod
from typing import Union

from rx.core.typing import Observable

from sfm2.interface.env_sensor_config_types import PressureSampleRate, TemperatureSamplingSetup
from sfm2.interface.sample_types import FloatSample


class Sfm2EnvSensors(ABC):
    """
    High-level interface for SFM2 env. sensors:
    - Temperature
    - Humidity
    - Pressure
    - Altitude
    """

    @property
    @abstractmethod
    def pressure(self) -> Observable[FloatSample]:
        pass

    @property
    @abstractmethod
    def altitude(self) -> Observable[FloatSample]:
        pass

    @property
    @abstractmethod
    def temperature(self) -> Observable[FloatSample]:
        pass

    @property
    @abstractmethod
    def humidity(self) -> Observable[FloatSample]:
        pass

    @property
    @abstractmethod
    def pressure_altitude_sample_rate(self) -> PressureSampleRate:
        """
        Reads or sets the sampling rate of pressure/altitude sensor.

        Cen be set to an int value representing rate in Hertz but the value must match one of available settings.
        If IMU sensors are disabled, setting this rate will also enable accelerometer.

        :type: Union[PressureSampleRate, int]
        :rtype: PressureSampleRate
        """
        pass

    @pressure_altitude_sample_rate.setter
    @abstractmethod
    def pressure_altitude_sample_rate(self, rate: Union[PressureSampleRate, int]) -> None:
        pass

    @property
    @abstractmethod
    def temperature_humidity_sampling_setup(self) -> TemperatureSamplingSetup:
        """
        Reads or sets the sampling setup of temperature/humidity sensor.

        See TemperatureSamplingSetup for a reference.

        :type: TemperatureSamplingSetup
        """
        pass

    @temperature_humidity_sampling_setup.setter
    @abstractmethod
    def temperature_humidity_sampling_setup(self, setup: TemperatureSamplingSetup) -> None:
        pass

    @abstractmethod
    def tare_altitude(self) -> None:
        """
        Tares altitude.

        Taring requires the pressure/altitude sampling rate to be configured (not 0).
        """
        pass
