from abc import ABC, abstractmethod
from typing import Union

from rx.core.typing import Observable

from sfm2.interface.sample_types import QuaternionSample, VectorSample, HeadingTiltSample
from sfm2.interface.sfm2_config import SampleRate


class Sfm2SensorFusion(ABC):
    """
    High-level interface for SFM2 Sensor Fusion
    """

    @property
    @abstractmethod
    def quaternion(self) -> Observable[QuaternionSample]:
        """
        Gets the quaternion output of Sensor Fusion (SFQT)

        :rtype: Observable[QuaternionSample]
        """
        pass

    @property
    @abstractmethod
    def untared_quaternion(self) -> Observable[QuaternionSample]:
        """
        Gets the untared quaternion output of Sensor Fusion (SFQ).
        This output differs from others in that it's not affected by taring.

        :rtype: Observable[QuaternionSample]
        """
        pass

    @property
    @abstractmethod
    def linear_acceleration(self) -> Observable[VectorSample]:
        """
        Gets the linear acceleration output of Sensor Fusion (SFLA).

        :rtype: Observable[VectorSample]
        """
        pass

    @property
    @abstractmethod
    def euler_angles(self) -> Observable[VectorSample]:
        """
        Gets the Euler angles output of Sensor Fusion (SFEA).

        :rtype: Observable[VectorSample]
        """
        pass

    @property
    @abstractmethod
    def heading_tilt(self) -> Observable[HeadingTiltSample]:
        """
        Gets the heading-tilt output of Sensor Fusion (SFCHT).

        :rtype: Observable[HeadingTiltSample]
        """
        pass

    @property
    @abstractmethod
    def calibrated_magnetometer(self) -> Observable[VectorSample]:
        """
        Gets the calibrated magnetometer output of Sensor Fusion (SFM).

        :rtype: Observable[VectorSample]
        """
        pass

    @property
    @abstractmethod
    def rate(self) -> SampleRate:
        """
        Reads or sets Sensor Fusion rate.

        Can be set to a float value, in this case the closest available setting will be applied.
        The library will check if the IMU sensors configuration is appropriate for the requested Sensor Fusion rate.
        If it's not, the IMU sensors will be configured first.

        :type: Union[SampleRate, float]
        :rtype: SampleRate
        """
        pass

    @rate.setter
    @abstractmethod
    def rate(self, rate: Union[SampleRate, float]) -> None:
        pass

    @property
    @abstractmethod
    def output_prescaler(self) -> int:
        pass

    @output_prescaler.setter
    @abstractmethod
    def output_prescaler(self, prescaler: int) -> None:
        pass


