from typing import Iterator

from sfm2._implementation.sfm2_discovery import discover_connected_sfm2s, connect_to_first_found_sfm2, \
    connect_to_all_found_sfm2s
from sfm2._implementation.sfm2_on_serial_port import Sfm2OnSerialPort
from sfm2.interface.exceptions import Sfm2Exception
from sfm2.interface.found_sfm2 import FoundSfm2
from sfm2.interface.sfm2 import Sfm2
from sfm2.interface.sfm2_init_options import Sfm2InitOptions, Sfm2CollectionInitOptions
from sfm2.utils.sfm2_collection import Sfm2Collection


class Sfm2Factory:
    """
    Static class used to create :class:`~sfm2.interface.sfm2.Sfm2` objects.

    See the static methods of this class for different connection options.
    """

    @staticmethod
    def create_on_serial_port(port_name: str, options: Sfm2InitOptions = Sfm2InitOptions()) -> Sfm2:
        """
        Connects to a SFM2 device on selected serial port

        :param port_name: Serial port name
        :type port_name: str
        :param options: Additional options. Can be omitted.
        :type options: Sfm2InitOptions
        :return: A connected SFM2
        :rtype: Sfm2
        """
        return Sfm2OnSerialPort(port_name, options)

    @staticmethod
    def find_all_on_serial_ports(options: Sfm2InitOptions = Sfm2InitOptions()) -> Iterator[FoundSfm2]:
        """
        Discovers all SFM2s connected to the computer.

        Calling this function does not yet connect to any device.
        Instead you get an iterable collection of found SFM2s,
        represented as :class:`~sfm2.interface.found_sfm2.FoundSfm2` objects, which you can then use to connect.

        :param options: Additional options. Can be omitted.
        :type options: Sfm2InitOptions
        :rtype: Iterator[FoundSfm2]
        """
        return discover_connected_sfm2s(options)

    @staticmethod
    def connect_first_found(options: Sfm2InitOptions = Sfm2InitOptions()) -> Sfm2:
        """
        Connects to first found SFM2

        :raises NoSfm2FoundException: When no SFM2s were found.
        :param options: Additional options. Can be omitted.
        :type options: Sfm2InitOptions
        :rtype: Sfm2
        """

        return connect_to_first_found_sfm2(options)

    @staticmethod
    def connect_all_found(device_options: Sfm2InitOptions = Sfm2InitOptions(),
                          collection_options: Sfm2CollectionInitOptions = Sfm2CollectionInitOptions()) -> Sfm2Collection:
        """
        Connects to all found SFM2

        :raises NoSfm2FoundException: When no SFM2s were found.
        :param options: Additional options. Can be omitted.
        :type options: Sfm2InitOptions
        :rtype: Sfm2
        """

        return connect_to_all_found_sfm2s(device_options, collection_options)

    def __init__(self):
        Sfm2Exception("Don't instantiate this class, use one of the static methods") \
            .raise_and_log_error()
