"""
Can be used to log SFM2 data to CSV files. See :func:`~sfm2.utils.csv_writer.write_to_csv`
"""
import logging
from dataclasses import dataclass
from datetime import timedelta
from typing import Iterable, Optional

from rx.core.abc import Disposable
from rx.core.typing import Observer, Observable
from rx.disposable import CompositeDisposable

from sfm2._implementation.internal_utils.observable_on_new_thread import observe_on_new_thread
from sfm2.interface.sample_types import Sample, SampleType, SampleFieldType

_log = logging.getLogger(__name__)


@dataclass(frozen=True)
class CsvWriterOptions:
    columns_separator: str = ","
    float_format: str = ".3f"
    add_header: bool = True
    include_units: bool = True
    skip_samples_missing_timestamp = True
    timestamp_column_header: str = "ts"
    overwrite_file_if_exists: bool = True
    create_new_thread: bool = True


def write_to_csv(data_observable: Observable[Sample], filepath: str,
                 options: CsvWriterOptions = CsvWriterOptions()) -> Disposable:
    """
    Start writing the data to CSV.

    :returns: A disposable that can be used to stop writing and close the CSV.
    :rtype: Disposable
    """
    writer = CsvWriter(filepath, options)
    observable = _create_observable(data_observable, options.create_new_thread)
    subscription = observable.subscribe(writer)
    return CompositeDisposable(subscription, writer)


class CsvWriter(Observer[Sample], Disposable):
    def __init__(self, filepath: str, options: CsvWriterOptions = CsvWriterOptions()):
        _log.info(f"Write CSV file {filepath}")
        _log.debug(f"Options: {options}")

        mode = "w" if options.overwrite_file_if_exists else "x"
        self._file = open(filepath, mode)
        self._should_write_header: bool = options.add_header
        self._options = options

    def on_next(self, value: Sample) -> None:
        if self._should_write_header:
            self._write_header(value.type)
            self._should_write_header = False

        self._write_sample_row(value)

    def on_error(self, error: Exception) -> None:
        _log.error(f"Error in data stream during CSV logging. Closing the file.\n{error}")
        self._close_file_if_not_already()

    def on_completed(self) -> None:
        _log.info("CSV file data stream completed")
        self._close_file_if_not_already()

    def dispose(self):
        _log.debug("Disposing")
        self._close_file_if_not_already()

    def _write_header(self, sample_type: SampleType) -> None:
        def format_column(field: SampleFieldType) -> str:
            if self._options.include_units:
                return self._column_header_with_unit(field.name, field.unit)
            else:
                return field.name

        ts_name = self._options.timestamp_column_header
        ts_full = self._column_header_with_unit(ts_name, "s") if self._options.include_units else ts_name

        value_columns_header = [format_column(f) for f in sample_type.fields]
        row = [ts_full] + value_columns_header
        self._write_row(row)

    @staticmethod
    def _column_header_with_unit(name: str, unit: str) -> str:
        return f"{name}[{unit}]"

    def _write_sample_row(self, sample: Sample) -> None:
        if sample.timestamp is None and self._options.skip_samples_missing_timestamp:
            return

        formatting = f"{{:{self._options.float_format}}}"
        values = [formatting.format(f) for f in sample.fields]

        row = [self._format_timestamp(sample.timestamp)] + values
        self._write_row(row)

    @staticmethod
    def _format_timestamp(ts: Optional[timedelta]) -> str:
        if ts is None:
            return ""

        return f"{ts.seconds}.{ts.microseconds:06d}"

    def _write_row(self, values: Iterable[str]) -> None:
        line = self._options.columns_separator.join(values)
        self._file.write(line + "\n")

    def _close_file_if_not_already(self):
        if not self._file.closed:
            _log.info("Closing CSV file")
            self._file.close()
        else:
            _log.debug("CSV file already closed")


def _create_observable(data_observable: Observable[Sample], create_new_thread: bool) -> Observable[Sample]:
    if create_new_thread:
        return observe_on_new_thread(data_observable)
    else:
        return data_observable
