import logging
from typing import Iterable, List, Union, Tuple, Iterator

from rx.core.abc import Disposable

from sfm2.interface.exceptions import DuplicateKeyException
from sfm2.interface.sfm2 import Sfm2
from sfm2.interface.sfm2_init_options import Sfm2CollectionInitOptions

_log = logging.getLogger(__name__)


class Sfm2Collection(Disposable):
    def __init__(self, sfm2s: Iterable[Sfm2], options: Sfm2CollectionInitOptions):
        self._devices = self._create_dict(sfm2s, options)
        self._options = options

    def __iter__(self) -> Iterator[Sfm2]:
        return (sfm2 for sfm2_list in self._devices.values() for sfm2 in sfm2_list)

    def __getitem__(self, key: str) -> Union[Sfm2, Tuple[Sfm2,]]:
        l = self._devices[key]
        if self._options.allow_duplicate_keys:
            return tuple(l)
        else:
            assert (len(l) == 1)
            return l[0]

    def __len__(self) -> int:
        return len(self._devices)

    def dispose(self):
        for d in self:
            try:
                d.dispose()
            except Exception:
                _log.exception("Could not dispose a device")

        self._devices.clear()

    @staticmethod
    def _create_dict(sfm2s: Iterable[Sfm2], options: Sfm2CollectionInitOptions):
        devices = dict()

        def get_key(s: Sfm2) -> str:
            return s.id if options.key_by_id_not_name else s.name

        def get_or_create_list(k: str) -> List:
            try:
                l = devices[k]
                if not options.allow_duplicate_keys:
                    DuplicateKeyException(f"Found duplicate device with key {k}."
                                          f"Disable this check by setting options.allow_duplicate_keys to True") \
                        .raise_and_log_error(_log)
                return l
            except KeyError:
                l = []
                devices[k] = l
                return l

        def add_to_dict(s: Sfm2) -> None:
            k = get_key(s)
            l = get_or_create_list(k)
            l.append(s)

        for s in sfm2s:
            add_to_dict(s)

        return devices
