"""
Can be used to calculate the actual incoming SFM2 data rates. See :func:`~sfm2.utils.throughput.calculate_throughput`
"""
from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from time import perf_counter_ns
from typing import Optional, Tuple, Union, Sequence

import rx
import rx.core.abc
import rx.operators as ops
from rx.core.typing import Observable, Disposable

from sfm2.interface.exceptions import UnexpectedDataException
from sfm2.interface.sample_types import Sample


@dataclass
class ThroughputInfo:
    samples_count: int
    time_period_ns: int

    min_device_ts_delta: timedelta
    max_device_ts_delta: timedelta

    @staticmethod
    def zero() -> ThroughputInfo:
        return ThroughputInfo(
            samples_count=0,
            time_period_ns=0,
            min_device_ts_delta=timedelta(),
            max_device_ts_delta=timedelta()
        )

    @property
    def rate_hz(self) -> float:
        if self.samples_count == 0:
            return 0

        time_period_s = self.time_period_ns * 1e-9
        return self.samples_count / time_period_s

    def __str__(self) -> str:
        if self.samples_count == 0:
            return "No data"

        rate_str = f"{self.rate_hz:.1f}Hz"
        if self.min_device_ts_delta != self.max_device_ts_delta:
            return f"{rate_str} (Inconsistent ts increments - some samples may be missing)"

        return rate_str


_DataStream = Observable[Sample]
_OneOrMoreDataStreams = Union[_DataStream, Sequence[_DataStream]]

_OneOrMoreInfoStreams = Union[Observable[ThroughputInfo], Observable[Tuple[ThroughputInfo,]]]


def calculate_throughput(data: _OneOrMoreDataStreams, interval_s: float = 1) -> _OneOrMoreInfoStreams:
    """Utility used to calculate the rate of SFM2 data.

    Transforms the passed data stream into a stream of throughput measurement results.
    Besides calculating data rate, the utility also examines timestamps to detect missing samples.

    If a collection of data streams is passed instead of a single data stream,
    then the result will be a single stream where each element is a tuple.
    Each element of that tuple is the throughput measurement result of its corresponding input data stream.

    The function won't have any effect until the returned stream is actually used.

    :param data: Data stream from the SFM2 or a sequence of data streams.
    :type data: Union[Observable[Sample], Sequence[Observable[Sample]]]
    :param interval_s: Interval at which the rate should be evaluated, in seconds. Defaults to 1 second.
    :type interval_s: float
    :returns: A stream of throughput results, update at the requested interval.
    :rtype: Union[Observable[ThroughputInfo], Observable[Tuple[ThroughputInfo,]]]
    """
    intervals = rx.timer(duetime=0, period=interval_s).pipe(
        ops.map(lambda _: None),
        ops.replay(buffer_size=1),
        ops.ref_count()
    )

    def create() -> Observable:
        if isinstance(data, rx.core.abc.Observable):
            return _calculate_throughput(data, intervals)
        else:
            throughputs = (_calculate_throughput(d, intervals) for d in data)
            return rx.zip(*throughputs)

    return create().pipe(
        ops.publish(),
        ops.ref_count()
    )


def _calculate_throughput(data: _DataStream, intervals: Observable) -> Observable[ThroughputInfo]:
    def create_observable(calculator: _Calculator) -> Observable[ThroughputInfo]:
        return data.pipe(
            ops.do_action(on_next=_verify_type_is_sample),
            ops.merge(intervals),
            ops.map(calculator.on_sample_or_interval),
            ops.filter(lambda info: info is not None)
        )

    return rx.using(lambda: _Calculator(), create_observable)


def _verify_type_is_sample(sample: Sample) -> None:
    if not isinstance(sample, Sample):
        UnexpectedDataException(f"Tried calculating throughput of {type(sample)} elements."
                                f" Work only with {Sample.__name__}").raise_and_log_error()


class _Calculator(Disposable):
    def __init__(self):
        self._current_interval_calculator: Optional[_IntervalCalculator] = None

    def on_sample_or_interval(self, sample_or_interval: Optional[Sample]) -> Optional[ThroughputInfo]:
        if sample_or_interval is not None:
            self._on_sample(sample_or_interval)
            return None
        else:
            return self._on_interval()

    def _on_sample(self, sample: Sample) -> None:
        if self._current_interval_calculator is not None:
            self._current_interval_calculator.on_sample(sample)

    def _on_interval(self) -> Optional[ThroughputInfo]:
        previous_calculator = self._current_interval_calculator
        self._current_interval_calculator = _IntervalCalculator()

        return previous_calculator.flush() \
            if previous_calculator is not None \
            else None

    def dispose(self):
        pass


class _IntervalCalculator:
    def __init__(self):
        self._samples_count: int = 0
        self._previous_device_ts: Optional[timedelta] = None
        self._device_ts_delta_range: Optional[Tuple[timedelta, timedelta]] = None
        self._first_sample_t_ns: Optional[int] = None
        self._last_sample_t_ns: Optional[int] = None

    def on_sample(self, sample: Sample) -> None:
        t_ns = perf_counter_ns()
        self._samples_count += 1

        self._update_device_ts_delta_range(sample.device_timestamp)

        if self._first_sample_t_ns is None:
            self._first_sample_t_ns = t_ns
        self._last_sample_t_ns = t_ns
        self._previous_device_ts = sample.device_timestamp

    def flush(self) -> ThroughputInfo:
        if self._samples_count < 2:
            return ThroughputInfo.zero()

        return ThroughputInfo(
            samples_count=self._samples_count - 1,
            time_period_ns=self._last_sample_t_ns - self._first_sample_t_ns,
            min_device_ts_delta=self._device_ts_delta_range[0],
            max_device_ts_delta=self._device_ts_delta_range[1]
        )

    def _update_device_ts_delta_range(self, device_ts: timedelta) -> None:
        if self._previous_device_ts is None:
            return

        delta = device_ts - self._previous_device_ts

        if self._device_ts_delta_range is None:
            self._device_ts_delta_range = delta, delta
            return

        range_min, range_max = self._device_ts_delta_range
        self._device_ts_delta_range = min(range_min, delta), max(range_max, delta)
