from __future__ import annotations

import logging
import math
from abc import abstractmethod
from dataclasses import dataclass
from datetime import timedelta, datetime
from time import perf_counter
from typing import Callable, Union, List

import rx
import rx.operators as ops
from rx.core.abc import Disposable
from rx.core.typing import Observable
from rx.disposable import CompositeDisposable
from rx.subject import ReplaySubject

from sfm2.interface.sfm2 import Sfm2
from sfm2.interface.sfm2_time import Sfm2Time
from .sfm2_collection import Sfm2Collection
from .._implementation.internal_utils.background_thread_executor import BackgroundThreadExecutor
from .._implementation.internal_utils.sfm2_rtc_time_conversion import sfm2_rtc_from_seconds, sfm2_rtc_to_timedelta

_log = logging.getLogger(__name__)


@dataclass(frozen=True)
class TimeSyncOptions:
    time_reports_rate_hz: float = 10
    time_reports_before_offset: int = 40
    max_allowed_offset_ticks: int = 30
    min_offsets_before_changing_sign: int = 3

    offsets_before_trim: int = 5
    trim_scale: float = 0.8


class TimeSyncStatus(Disposable):
    """
    Represents the status of an ongoing synchronization.

    Disposing it stops the synchronization.
    """

    @abstractmethod
    def wait_for_sync(self, timeout: timedelta) -> None:
        """
        Blocks until the clocks are properly synchronized.

        :param timeout: Maximum time to wait for the synchronization.
        """
        pass

    @property
    @abstractmethod
    def offset(self) -> Observable[timedelta]:
        pass


class TimeSync(Disposable):
    """
    Utility used to synchronize SFM2 clocks to the PC time.

    Creating a TimeSync starts a 'PC clock', that measures time elapsed on PC.
    """

    def __init__(self, options: TimeSyncOptions = TimeSyncOptions()):
        self._zero_timestamp = datetime.now()
        _log.info(f"Starting Sync at {self._zero_timestamp}")

        self._options = options
        self._utils = _Utils()
        self._disposables = CompositeDisposable()

    def sync(self, sfm2_or_collection: Union[Sfm2, Sfm2Collection]) -> TimeSyncStatus:
        """
        Starts synchronizing clocks on one or more SFM2s.

        This method can be called multiple times at any time.
        The SFM2 clock will always be synchronized to the same time (with t0 at zero_timestamp).

        :param sfm2_or_collection: One sfm2 or a collection.
        :type sfm2_or_collection: Union[Sfm2, Sfm2Collection]

        :return: object representing the status of the synchronization
        :rtype: TimeSyncStatus
        """
        if isinstance(sfm2_or_collection, Sfm2Collection):
            all_statuses = list()
            for sfm2 in sfm2_or_collection:
                s = self._sync_sfm2_impl(sfm2)
                all_statuses.append(s)
            return _CompositeTimeSyncStatus(all_statuses, self._options)
        else:
            return self._sync_sfm2_impl(sfm2_or_collection)

    @property
    def zero_timestamp(self) -> datetime:
        """
        Returns system time corresponding to t=0.

        :rtype: datetime
        """
        return self._zero_timestamp

    def get_time(self) -> timedelta:
        """
        Returns current clock reading.

        Can be used to synchronize other events to the SFM2 measurements.

        :rtype: timedelta
        """
        seconds = self._utils.get_pc_time_s()
        return timedelta(seconds=seconds)

    def _sync_sfm2_impl(self, sfm2: Sfm2) -> TimeSyncStatus:
        pc_time = timedelta(seconds=self._utils.get_pc_time_s())
        sfm2.time.set(pc_time)
        sync = _Sync(sfm2.time, self._utils, self._options)
        self._disposables.add(sync)
        return sync

    def dispose(self):
        self._disposables.dispose()
        self._utils.dispose()


class _Sync(TimeSyncStatus):
    def __init__(self, sfm2_time: Sfm2Time,
                 utils: _Utils,
                 options: TimeSyncOptions):
        self._sfm2_time = sfm2_time
        self._utils = utils
        self._options = options
        self._trim = _Trim(sfm2_time, utils, options)
        self._offset_subject = ReplaySubject(buffer_size=1)

        self._last_offset_calculated: int = 0
        self._offsets_since_sign_changed = options.min_offsets_before_changing_sign

        self._reports_count = 0
        self._max_time_diff_s = float("-inf")

        self._disposable = sfm2_time.reports.subscribe(self._process_time_report)
        sfm2_time.reports_rate_hz = options.time_reports_rate_hz

    def wait_for_sync(self, timeout: timedelta = timedelta.max) -> None:
        max_allowed = sfm2_rtc_to_timedelta(self._options.max_allowed_offset_ticks)
        self._offset_subject.pipe(
            ops.first(lambda o: abs(o) <= max_allowed),
            ops.timeout(timeout.total_seconds())
        ).run()

    @property
    def offset(self) -> Observable[timedelta]:
        return self._offset_subject

    def dispose(self):
        self._sfm2_time.reports_rate_hz = 0
        self._offset_subject.dispose()
        self._disposable.dispose()

    def _process_time_report(self, sfm2_time: timedelta) -> None:
        self._update_max_time_diff(sfm2_time)
        self._reports_count += 1

        if self._reports_count >= self._options.time_reports_before_offset:
            offset_ticks = -sfm2_rtc_from_seconds(self._max_time_diff_s)
            self._process_calculated_offset(offset_ticks)

            self._max_time_diff_s = float("-inf")
            self._reports_count = 0

    def _process_calculated_offset(self, offset: int) -> None:
        _log.debug(f"Offset: {offset}")
        self._offset_subject.on_next(sfm2_rtc_to_timedelta(offset))

        if math.fabs(offset) <= self._options.max_allowed_offset_ticks:
            _log.debug(f"{offset} is within max allowed {self._options.max_allowed_offset_ticks}")
            return

        if self._check_sign(offset):
            self._apply_offset(offset)
        else:
            _log.debug("Offset sign failed sign test")

    def _check_sign(self, offset: int) -> bool:
        sign_changed = offset * self._last_offset_calculated < 0
        self._last_offset_calculated = offset

        if sign_changed:
            self._offsets_since_sign_changed = 0
        else:
            self._offsets_since_sign_changed += 1

        return self._offsets_since_sign_changed > self._options.min_offsets_before_changing_sign

    def _apply_offset(self, offset: int) -> None:
        self._trim.on_offset(offset)
        self._last_offset_calculated = offset
        self._utils.execute_in_background(lambda: self._sfm2_time.apply_offset(offset))

    def _update_max_time_diff(self, sfm2_time: timedelta) -> None:
        pc_time_s = self._utils.get_pc_time_s()
        time_diff_s = sfm2_time.total_seconds() - pc_time_s
        if time_diff_s > self._max_time_diff_s:
            self._max_time_diff_s = time_diff_s


class _Trim:
    def __init__(self, sfm2_time: Sfm2Time, utils: _Utils, options: TimeSyncOptions):
        self._sfm2_time = sfm2_time
        self._utils = utils
        self._options = options

        self._should_trim = False
        self._offsets_count = -1
        self._first_offset_pc_time_s = math.nan
        self._cumulative_offset = 0

    def on_offset(self, offset: int) -> None:
        t = self._utils.get_pc_time_s()

        if self._offsets_count < 0:
            self._first_offset_pc_time_s = t
            self._offsets_count = 0
            return

        self._cumulative_offset += offset
        self._offsets_count += 1

        if self._offsets_count >= self._options.offsets_before_trim:
            if self._should_trim:
                self._process_trim(t)
            self._should_trim = not self._should_trim

    def _process_trim(self, last_offset_pc_time_s: float) -> None:
        time_passed_on_pc_s = last_offset_pc_time_s - self._first_offset_pc_time_s
        time_passed_on_pc = timedelta(seconds=time_passed_on_pc_s)
        cumulative_offset_scaled = self._cumulative_offset * self._options.trim_scale
        time_added_with_offset_scaled = sfm2_rtc_to_timedelta(cumulative_offset_scaled)
        time_passed_on_sfm2 = time_passed_on_pc - time_added_with_offset_scaled
        self._apply_trim(time_passed_on_sfm2, time_passed_on_pc)
        self._reset_state()

    def _apply_trim(self, time_passed_on_sfm2: timedelta, time_passed_on_pc: timedelta) -> None:
        def calibrate() -> None:
            self._sfm2_time.calibrate(time_passed_on_sfm2=time_passed_on_sfm2,
                                      time_passed_on_pc=time_passed_on_pc)

        self._utils.execute_in_background(calibrate)

    def _reset_state(self) -> None:
        self._offsets_count = -1
        self._first_offset_pc_time_s = math.nan
        self._cumulative_offset = 0


class _Utils(Disposable):
    def __init__(self):
        self._time_origin_s = perf_counter()
        self._executor = BackgroundThreadExecutor()

    def get_pc_time_s(self) -> float:
        return perf_counter() - self._time_origin_s

    def execute_in_background(self, action: Callable[[], None]):
        self._executor.execute(action)

    def dispose(self):
        self._executor.dispose()


class _CompositeTimeSyncStatus(TimeSyncStatus):
    def __init__(self, statuses: List[TimeSyncStatus], options: TimeSyncOptions):
        self._statuses = statuses
        self._options = options

        offsets = (s.offset for s in statuses)
        self._offset = rx.combine_latest(*offsets).pipe(
            ops.map(self._find_max_offset),
            ops.replay(buffer_size=1),
            ops.ref_count()
        )

    def wait_for_sync(self, timeout: timedelta) -> None:
        max_allowed = sfm2_rtc_to_timedelta(self._options.max_allowed_offset_ticks)
        self._offset.pipe(
            ops.first(lambda o: abs(o) <= max_allowed),
            ops.timeout(timeout.total_seconds())
        ).run()

    @property
    def offset(self) -> Observable[timedelta]:
        return self._offset

    def dispose(self):
        for s in self._statuses:
            s.dispose()

    @staticmethod
    def _find_max_offset(offsets: List[timedelta]) -> timedelta:
        max_abs_s = float("-inf")
        found_offset = None

        for o in offsets:
            abs_s = abs(o.total_seconds())
            if abs_s > max_abs_s:
                max_abs_s = abs_s
                found_offset = o
        return found_offset
